#define _XOPEN_SOURCE 500

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "def.h"
#include "game.h"
#include "gint/gint.h"

#define G_HEIGHT_SCALER			0.05f
#define G_GRAVITY			-24.0f

#define G_QU(x)				((x)*(x))

#define G_INDEX(n,p)			(*(ui16*)((uint)p+GI_MESH_ISIZE*n))
#define G_VERT_V(n,p,i)			(*(f32*)((uint)p+GI_MESH_VSIZE*n+4*i))
#define G_VERT_T(n,p,i)			(*(f32*)((uint)p+GI_MESH_VSIZE*n+12+4*i))
#define G_VERT_N(n,p,i)			(*(ui8*)((uint)p+GI_MESH_VSIZE*n+20+i))
#define G_COLOR(n,p,r,g,b)		{ \
	G_VERT_N((n),(p),0)=r; \
	G_VERT_N((n),(p),1)=g; \
	G_VERT_N((n),(p),2)=b; \
	G_VERT_N((n),(p),3)=255; }
#define G_COLOR_MONO(n,p,c)		G_COLOR((n),(p),(c),(c),(c))
#define G_COLOR_CLAMP(cf)		((cf)>0.25f?0.25f:(cf))

g_world *g_world_create(	gi_font *font,
				const char *share,
				int map )
{
	f32 cfog[][4] = {	{ 0.5f, 0.7f, 0.8f, 1.0f },
       				{ 0.75f, 0.75f, 0.7f, 1.0f },
				{ 0.5f, 0.5f, 0.5f, 1.0f } } ;

	glFogfv( GL_FOG_COLOR, cfog[ map ] );
	glClearColor( cfog[ map ][ 0 ], cfog[ map ][ 1 ], cfog[ map ][ 2 ], cfog[ map ][ 3 ] );

	char buf[ 256 ];
	sprintf( buf, "%smap%i.png", share, map );

	g_world *world = malloc( sizeof( g_world ) );
	world->dta = 0.0;

	gi_img_load_png( buf, &world->hmap );

	world->mesh_map = malloc( sizeof( gi_mesh* ) * world->hmap.width * world->hmap.height / 64 );
	world->map_w = world->hmap.width / 8;
	world->map_h = world->hmap.height / 8;

	void *idata = malloc( 6 * 64 * GI_MESH_ISIZE );
	void *vdata = malloc( 4 * 64 * GI_MESH_VSIZE );

	int _i, _j;
	for ( _i = 0; _i < world->map_w; _i++ )
		for ( _j = 0; _j < world->map_h; _j++ )
	{
		int sdi = _i == 0 ? 1 : 0 ;
		int sdj = _j == 0 ? 1 : 0 ;
		int adi = _i < ( world->map_w - 1 ) ? 8 : 6 ;
		int adj = _j < ( world->map_h - 1 ) ? 8 : 6 ;
		int i, j, c = 0, ic = 0 ;
		for ( i = ( _i*8 + sdi ); i < ( _i*8 + adi ); i++ )
			for ( j = ( _j*8 + sdj ); j < ( _j*8 + adj ); j++ )
		{
			f32 h0 = G_HEIGHT_SCALER *
				( f32 ) world->hmap.raw[ 3 * ( j + i * world->hmap.height ) ];
			f32 h1 = G_HEIGHT_SCALER *
				( f32 ) world->hmap.raw[ 3 * ( ( j + 1 ) + i * world->hmap.height ) ];
			f32 h1_1 = G_HEIGHT_SCALER *
				( f32 ) world->hmap.raw[ 3 * ( ( j - 1 ) + i * world->hmap.height ) ];
			f32 h1_2 = G_HEIGHT_SCALER *
				( f32 ) world->hmap.raw[ 3 * ( ( j + 2 ) + i * world->hmap.height ) ];
			f32 h1_3 = G_HEIGHT_SCALER *
				( f32 ) world->hmap.raw[ 3 * ( ( j + 1 ) + ( i - 1 ) * world->hmap.height ) ];
			f32 h2 = G_HEIGHT_SCALER *
				( f32 ) world->hmap.raw[ 3 * ( j + ( i + 1 ) * world->hmap.height ) ];
			f32 h2_1 = G_HEIGHT_SCALER *
				( f32 ) world->hmap.raw[ 3 * ( j + ( i - 1 ) * world->hmap.height ) ];
			f32 h2_2 = G_HEIGHT_SCALER *
				( f32 ) world->hmap.raw[ 3 * ( j + ( i + 2 ) * world->hmap.height ) ];
			f32 h2_3 = G_HEIGHT_SCALER *
				( f32 ) world->hmap.raw[ 3 * ( ( j - 1 ) + ( i + 1 ) * world->hmap.height ) ];
			f32 h3 = G_HEIGHT_SCALER *
				( f32 ) world->hmap.raw[ 3 * ( ( j + 1 ) + ( i + 1 ) * world->hmap.height ) ];
			f32 h3_1 = G_HEIGHT_SCALER *
				( f32 ) world->hmap.raw[ 3 * ( ( j + 2 ) + ( i + 1 ) * world->hmap.height ) ];
			f32 h3_2 = G_HEIGHT_SCALER *
				( f32 ) world->hmap.raw[ 3 * ( ( j + 1 ) + ( i + 2 ) * world->hmap.height ) ];
			f32 c0 = 255 - ( ui8 ) ( 255.0f * G_COLOR_CLAMP( sqrt( G_QU( fabs( h1 - h1_1 ) ) + G_QU( fabs( h2 - h2_1 ) ) ) * 0.25f ) );
			f32 c1 = 255 - ( ui8 ) ( 255.0f * G_COLOR_CLAMP( sqrt( G_QU( fabs( h0 - h1_2 ) ) + G_QU( fabs( h3 - h1_3 ) ) ) * 0.25f ) );
			f32 c2 = 255 - ( ui8 ) ( 255.0f * G_COLOR_CLAMP( sqrt( G_QU( fabs( h0 - h2_2 ) ) + G_QU( fabs( h3 - h2_3 ) ) ) * 0.25f ) );
			f32 c3 = 255 - ( ui8 ) ( 255.0f * G_COLOR_CLAMP( sqrt( G_QU( fabs( h2 - h3_1 ) ) + G_QU( fabs( h1 - h3_2 ) ) ) * 0.25f ) );
			G_VERT_V( c, vdata, 0 ) = ( f32 ) i;
			G_VERT_V( c, vdata, 1 ) = h0;
			G_VERT_V( c, vdata, 2 ) = ( f32 ) j;
			G_VERT_T( c, vdata, 0 ) = 0.125f * ( f32 ) ( i );
			G_VERT_T( c, vdata, 1 ) = 0.125f * ( f32 ) ( j );
			G_COLOR_MONO( c, vdata, c0 );
			G_INDEX( ic, idata ) = c; ic++; c++;
			G_VERT_V( c, vdata, 0 ) = ( f32 ) i;
			G_VERT_V( c, vdata, 1 ) = h1;
			G_VERT_V( c, vdata, 2 ) = ( f32 ) j + 1;
			G_VERT_T( c, vdata, 0 ) = 0.125f * ( f32 ) ( i );
			G_VERT_T( c, vdata, 1 ) = 0.125f * ( f32 ) ( j + 1 );
			G_COLOR_MONO( c, vdata, c1 );
			G_INDEX( ic, idata ) = c; ic++; c++;
			G_VERT_V( c, vdata, 0 ) = ( f32 ) i + 1;
			G_VERT_V( c, vdata, 1 ) = h2;
			G_VERT_V( c, vdata, 2 ) = ( f32 ) j;
			G_VERT_T( c, vdata, 0 ) = 0.125f * ( f32 ) ( i + 1 );
			G_VERT_T( c, vdata, 1 ) = 0.125f * ( f32 ) ( j );
			G_COLOR_MONO( c, vdata, c2 );
			G_INDEX( ic, idata ) = c; ic++; c++;

			G_INDEX( ic, idata ) = c - 1; ic++;
			G_INDEX( ic, idata ) = c - 2; ic++;
			G_VERT_V( c, vdata, 0 ) = ( f32 ) i + 1;
			G_VERT_V( c, vdata, 1 ) = h3;
			G_VERT_V( c, vdata, 2 ) = ( f32 ) j + 1;
			G_VERT_T( c, vdata, 0 ) = 0.125f * ( f32 ) ( i + 1 );
			G_VERT_T( c, vdata, 1 ) = 0.125f * ( f32 ) ( j + 1 );
			G_COLOR_MONO( c, vdata, c3 );
			G_INDEX( ic, idata ) = c; ic++; c++;
		}

		world->mesh_map[ _j + _i * world->map_h ] = gi_mesh_create( c, ic, vdata, idata );
		world->mesh_map[ _j + _i * world->map_h ]->light = TRUE;
	}

	free( idata );
	free( vdata );

	sprintf( buf, "%smap%i_tex.png", share, map );
	world->texture = gi_img_load_png( buf, NULL );

	world->cam[ 0 ] = 0.0f;
	world->cam[ 1 ] = 64.0f;
	world->cam[ 2 ] = 0.0f;

	sprintf( buf, "%ssky.mesh", share );
	world->sky_mesh = gi_mesh_load( buf );

	sprintf( buf, "%ssky.png", share );
	world->sky_tex = gi_img_load_png( buf, NULL );

	sprintf( buf, "%splane.mesh", share );
	world->plane_mesh = gi_mesh_load( buf );

	sprintf( buf, "%sflare.png", share );
	world->flare_tex = gi_img_load_png( buf, NULL );

	sprintf( buf, "%sball.mesh", share );
	world->ball.mesh = gi_mesh_load( buf );

	sprintf( buf, "%sgun.mesh", share );
	world->ball.mesh_gun = gi_mesh_load( buf );

	sprintf( buf, "%sball_tex.png", share );
	world->ball.texture = gi_img_load_png( buf, NULL );

	i8 ldir[] = { 0, 127, 0 };
	gi_mesh_mklight( world->ball.mesh_gun, ldir, 64 );

	world->sky_pos = 0.0f;

	world->ball.pos[ 0 ] = 16.0f;
	world->ball.pos[ 1 ] = 0.0f;
	world->ball.pos[ 2 ] = 16.0f;

	world->ball.ang[ 0 ] = 0.0f;
	world->ball.ang[ 1 ] = 0.0f;
	world->ball.ang[ 2 ] = 0.0f;

	world->ball.jmp_h = 0.0f;
	world->ball.jmp_v = 0.0f;
	world->ball.gun_a = 0.0f;
	world->ball.gun_ang = 0.0f;
	world->ball.gun_time = 0.0f;
	world->ball.gun_mode = 0;

	world->ball.fpm = TRUE;

	world->win = FALSE;
	world->gstr_win = gi_string_create( font, 0.1f, _("!!! YOU WIN !!!") );
	world->gstr_win->fcl[ 0 ] = 255;
	world->gstr_win->fcl[ 1 ] = 32;
	world->gstr_win->fcl[ 2 ] = 32;
	world->gstr_win->fcl[ 3 ] = 200;

	gi_get( GI_WINDOW_WIDTH, &world->cetr_mx );
	gi_get( GI_WINDOW_HEIGHT, &world->cetr_my );
	world->cetr_mx /= 2;
	world->cetr_my /= 2;

	return world;
}

void g_world_delete( g_world *world )
{
	glClearColor( 0.0f, 0.0f, 0.0f, 1.0f );

	int _i, _j;
	for ( _i = 0; _i < world->map_w; _i++ )
		for ( _j = 0; _j < world->map_h; _j++ )
			gi_mesh_delete( world->mesh_map[ _j + _i * world->map_h ] );
	free( world->mesh_map );
	gi_img_delete( 0, &world->hmap );
	gi_img_delete( world->texture, NULL );
	gi_img_delete( world->sky_tex, NULL );
	gi_img_delete( world->flare_tex, NULL );
	gi_img_delete( world->ball.texture, NULL );
	gi_mesh_delete( world->ball.mesh );
	gi_mesh_delete( world->ball.mesh_gun );
	gi_mesh_delete( world->sky_mesh );
	gi_mesh_delete( world->plane_mesh );
	gi_string_delete( world->gstr_win );
	free( world );
}

void g_world_update(	g_world *world,
			gi_input *input,
			f64 dt )
{
	world->dta += dt;
	dt = 0.01;

	while ( dt < world->dta )
	{
		if ( world->ball.gun_mode > 0 )
		{
			world->ball.gun_time -= dt;
			if ( world->ball.gun_time < 0.0f )
			{
				world->ball.gun_mode++;
				world->ball.gun_time = 0.04f;
			}

			if ( world->ball.gun_mode > 2 )
				world->ball.gun_mode = 0;
		}

		if ( world->win == FALSE )
		{
			if ( input->keys[ GI_W ] == GI_KEY_DOWN )
			{
				world->ball.pos[ 0 ] += 12.0f * cosf( world->ball.ang[ 1 ] ) * dt;
				world->ball.pos[ 2 ] += 12.0f * sinf( world->ball.ang[ 1 ] ) * dt;
				world->ball.ang[ 0 ] += ( 12.0f * dt ) / ( 2.0f ) ;
				world->ball.gun_a += 18.0f * dt;
			}

			if ( input->keys[ GI_S ] == GI_KEY_DOWN )
			{
				world->ball.pos[ 0 ] -= 12.0f * cosf( world->ball.ang[ 1 ] ) * dt;
				world->ball.pos[ 2 ] -= 12.0f * sinf( world->ball.ang[ 1 ] ) * dt;
				world->ball.ang[ 0 ] -= ( 12.0f * dt ) / ( 2.0f ) ;
				world->ball.gun_a += 18.0f * dt;
			}

			if ( ( input->keys[ GI_SPACE ] == GI_KEY_DOWN ) &&
			   	( world->ball.jmp_h == 0.0f ) )
			{
				world->ball.jmp_v = 8.0f;
				input->keys[ GI_S ] = GI_KEY_OFF;
			}

			if ( world->ball.fpm )
			{
				if ( input->keys[ GI_A ] == GI_KEY_DOWN )
				{
					world->ball.pos[ 0 ] += 12.0f * sinf( world->ball.ang[ 1 ] ) * dt;
					world->ball.pos[ 2 ] += -12.0f * cosf( world->ball.ang[ 1 ] ) * dt;
					world->ball.gun_a += 18.0f * dt;
				}

				if ( input->keys[ GI_D ] == GI_KEY_DOWN )
				{
					world->ball.pos[ 0 ] += -12.0f * sinf( world->ball.ang[ 1 ] ) * dt;
					world->ball.pos[ 2 ] += 12.0f * cosf( world->ball.ang[ 1 ] ) * dt;
					world->ball.gun_a += 18.0f * dt;
				}

				world->ball.ang[ 1 ] += 0.25f * ( f64 ) ( input->mouse_x - world->cetr_mx ) * dt;
				world->ball.ang[ 2 ] += - 0.25f * ( f64 ) ( input->mouse_y - world->cetr_my ) * dt;
				world->ball.ang[ 2 ] = world->ball.ang[ 2 ] < ( - M_PI * 0.499f ) ? ( - M_PI * 0.499f ) : world->ball.ang[ 2 ];
				world->ball.ang[ 2 ] = world->ball.ang[ 2 ] > ( M_PI * 0.499f ) ? ( M_PI * 0.499f ) : world->ball.ang[ 2 ];

				input->mouse_x = world->cetr_mx;
				input->mouse_y = world->cetr_my;
			}
			else
			{
				if ( input->keys[ GI_A ] == GI_KEY_DOWN )
					world->ball.ang[ 1 ] -= 4.0f * dt;
				if ( input->keys[ GI_D ] == GI_KEY_DOWN )
					world->ball.ang[ 1 ] += 4.0f * dt;
			}

			if ( input->keys[ GI_BUTTON_LEFT ] == GI_KEY_DOWN )
			{
				if ( world->ball.gun_mode == 0 )
				{
					world->ball.gun_time = 0.04f;
					world->ball.gun_mode = 1;
					world->ball.gun_ang = 360.0f * ( f32 ) rand() / ( f32 ) RAND_MAX;

					/*f32 beg[] = { world->ball.pos[ 0 ],
				       			world->ball.pos[ 1 ] + 1.0f,
							world->ball.pos[ 2 ] };
					f32 ray[] = { cosf( world->ball.ang[ 1 ] ) * cosf( world->ball.ang[ 2 ] ),
							sinf( world->ball.ang[ 2 ] ) + 1.0f,
							sinf( world->ball.ang[ 1 ] ) * cosf( world->ball.ang[ 2 ] ) };
					f32 ret[ 3 ];

					g_world_get_point( world, beg, ray, ret );

					int x = ( int ) beg[ 0 ];
					int y = ( int ) beg[ 2 ];
					printf( "%i %i \n ", x, y );
					ui8 data[] = { 0, 0, 0, 255 };
					glBindTexture( GL_TEXTURE_2D, world->texture );
					glTexSubImage2D( GL_TEXTURE_2D, 0, 0, 0, 1, 1,
							GL_RGBA, GL_UNSIGNED_BYTE, data );
					printf( "err=%x \n", glGetError() );*/
				}
			}

			if ( input->keys[ GI_F ] == GI_KEY_DOWN )
			{
				world->ball.fpm = !world->ball.fpm;
				/*if ( world->ball.fpm )
					world->ball.ang[ 2 ] = 0.0f;*/
				input->keys[ GI_F ] = GI_KEY_OFF;
			}
		}

		world->ball.jmp_v += G_GRAVITY * dt;
		world->ball.jmp_h += world->ball.jmp_v * dt;
		if ( world->ball.jmp_h < 0.0f )
		{
			world->ball.jmp_h = 0.0f;
			world->ball.jmp_v = 0.0f;
		}

		world->ball.pos[ 1 ] = g_world_get_h( world,
				world->ball.pos[ 0 ],
				world->ball.pos[ 2 ] ) + world->ball.jmp_h;

		int i, j ;
		for ( i = -1; i <= 1; i++ )
			for ( j = -1; j <= 1; j++ )
		{
			int pos_i = ( int ) world->ball.pos[ 0 ];
			int pos_j = ( int ) world->ball.pos[ 2 ];

			ui8 p0 = world->hmap.raw[ 3 * ( ( pos_j + j ) + ( pos_i + i ) * world->hmap.height ) ];
			ui8 p1 = world->hmap.raw[ 3 * ( ( 1 + pos_j + j ) + ( pos_i + i ) * world->hmap.height ) ];
			ui8 p2 = world->hmap.raw[ 3 * ( ( pos_j + j ) + ( 1 + pos_i + i ) * world->hmap.height ) ];
			ui8 p3 = world->hmap.raw[ 3 * ( ( 1 + pos_j + j ) + ( 1 + pos_i + i ) * world->hmap.height ) ];

			if ( ( i == 0 && j == 0 ) &&
					( p0 == 0 && p1 == 0 && p2 == 0 && p3 == 0 ) )
				world->win = TRUE;

			if ( p0 == 255 || p1 == 255 || p2 == 255 || p3 == 255 )
			{
				int rec_i = pos_i + i;
				int rec_j = pos_j + j;
				f32 sub_x = ( ( f32 ) rec_i + 0.5f ) - world->ball.pos[ 0 ];
				f32 sub_y = ( ( f32 ) rec_j + 0.5f ) - world->ball.pos[ 2 ];

				if ( ( fabs( sub_x ) < 1.5f ) &&
					( fabs( sub_y ) < 1.5f ) )
				{
					if ( fabs( sub_x ) > fabs( sub_y ) )
					{
						if ( sub_x < 0.0f )
							world->ball.pos[ 0 ] += ( sub_x + 1.5f );
						else
							world->ball.pos[ 0 ] += ( sub_x - 1.5f );
					}
					else
					{
						if ( sub_y < 0.0f )
							world->ball.pos[ 2 ] += ( sub_y + 1.5f );
						else
							world->ball.pos[ 2 ] += ( sub_y - 1.5f );
					}
				}
			}
		}

		if ( world->ball.fpm )
		{
			world->cam[ 0 ] = world->ball.pos[ 0 ];
			world->cam[ 1 ] = world->ball.pos[ 1 ] + 1.0f;
			world->cam[ 2 ] = world->ball.pos[ 2 ];

			world->ball.tgt[ 0 ] = world->ball.pos[ 0 ] +
				cosf( world->ball.ang[ 1 ] ) * cosf( world->ball.ang[ 2 ] );
			world->ball.tgt[ 2 ] = world->ball.pos[ 2 ] +
				sinf( world->ball.ang[ 1 ] ) * cosf( world->ball.ang[ 2 ] );
			world->ball.tgt[ 1 ] = world->ball.pos[ 1 ] +
				sinf( world->ball.ang[ 2 ] ) + 1.0f;
		}
		else
		{
			f32 camto[] = { world->ball.pos[ 0 ] - 8.0f * cosf( world->ball.ang[ 1 ] ),
				world->ball.pos[ 1 ] + 12.0f,
				world->ball.pos[ 2 ] - 8.0f * sinf( world->ball.ang[ 1 ] ) };

			world->cam[ 0 ] += 2.0f * ( camto[ 0 ] - world->cam[ 0 ] ) * dt;
			world->cam[ 1 ] += 2.0f * ( camto[ 1 ] - world->cam[ 1 ] ) * dt;
			world->cam[ 2 ] += 2.0f * ( camto[ 2 ] - world->cam[ 2 ] ) * dt;
		}

		world->sky_pos += 0.02f * dt;

		world->dta -= dt;
	}
}

void g_world_draw( g_world *world )
{
	f32 up[] = { 0.0f, 1.0f, 0.0f };
	if ( world->ball.fpm )
		gi_lookat( world->cam, world->ball.tgt, up );
	else
		gi_lookat( world->cam, world->ball.pos, up );

	/* sky */
	glPushMatrix();
	glTranslatef( world->cam[ 0 ], world->cam[ 1 ], world->cam[ 2 ] );

	glMatrixMode( GL_TEXTURE );
	glPushMatrix();
	glTranslatef( world->sky_pos, world->sky_pos, 0.0f );
	glScalef( 4.0f, 4.0f, 1.0f );

	glDisable( GL_DEPTH_TEST );
	glDepthMask( GL_FALSE );
	glEnable( GL_BLEND );
	glBlendFunc( GL_SRC_COLOR, GL_ONE_MINUS_SRC_COLOR );
	glColor4f( 1.0f, 1.0f, 1.0f, 1.0f );

	glBindTexture( GL_TEXTURE_2D, world->sky_tex );
	gi_mesh_draw( world->sky_mesh );

	glDisable( GL_BLEND );
	glEnable( GL_DEPTH_TEST );
	glDepthMask( GL_TRUE );

	glPopMatrix();
	glMatrixMode( GL_MODELVIEW );

	glPopMatrix();

	/* main */
	glBindTexture( GL_TEXTURE_2D, world->texture );

	int _i, _j;
	int i_beg = ( int ) ( world->ball.pos[ 0 ] * 0.125f ) - 5;
	int j_beg = ( int ) ( world->ball.pos[ 2 ] * 0.125f ) - 5;
	int i_end = ( int ) ( world->ball.pos[ 0 ] * 0.125f ) + 5;
	int j_end = ( int ) ( world->ball.pos[ 2 ] * 0.125f ) + 5;

	i_beg = i_beg < 0 ? 0 : i_beg;
	j_beg = j_beg < 0 ? 0 : j_beg;
	i_end = i_end > world->map_w ? world->map_w : i_end;
	j_end = j_end > world->map_h ? world->map_h : j_end;

	for ( _i = i_beg; _i < i_end; _i++ )
		for ( _j = j_beg; _j < j_end; _j++ )
			gi_mesh_draw( world->mesh_map[ _j + _i * world->map_h ] );

	glColor4f( 1.0f, 1.0f, 1.0f, 1.0f );

	/* player */
	if ( !world->ball.fpm )
	{
		glPushMatrix();
		glTranslatef(	world->ball.pos[ 0 ],
				world->ball.pos[ 1 ] + 1.0f,
				world->ball.pos[ 2 ] );

		glRotatef( - world->ball.ang[ 1 ] * ( 180.0f / M_PI ), 0.0f, 1.0f, 0.0f );
		glRotatef( - world->ball.ang[ 0 ] * ( 180.0f / M_PI ), 0.0f, 0.0f, 1.0f );

		glBindTexture( GL_TEXTURE_2D, world->ball.texture );
		gi_mesh_draw( world->ball.mesh );

		glPopMatrix();
	}

	/* player gun */

	glPushMatrix();
	glTranslatef(	world->ball.pos[ 0 ],
			world->ball.pos[ 1 ] + 1.0f,
			world->ball.pos[ 2 ] );

	glRotatef( - world->ball.ang[ 1 ] * ( 180.0f / M_PI ) + 1.0f * sinf( world->ball.gun_a * 0.5f ), 0.0f, 1.0f, 0.0f );
	glRotatef( + world->ball.ang[ 2 ] * ( 180.0f / M_PI ) * 1.05f + 1.0f * sinf( world->ball.gun_a ), 0.0f, 0.0f, 1.0f );
	glTranslatef( ( world->ball.gun_mode == 1 ) ? -0.04f : 0.0f, 0.0f, 0.0f );

	glBindTexture( GL_TEXTURE_2D, world->ball.texture );
	gi_mesh_draw( world->ball.mesh_gun );

	if ( world->ball.gun_mode == 1 )
	{
		glRotatef( -90.0f, 0.0f, 1.0f, 0.0f );
		glTranslatef( 0.28f, -0.45f, -1.35f );
		glScalef( 0.4f, 0.4f, 0.4f );
		glRotatef( world->ball.gun_ang, 0.0f, 0.0f, 1.0f );

		glEnable( GL_BLEND );
		glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );

		glBindTexture( GL_TEXTURE_2D, world->flare_tex );
		glColor4f( 1.0f, 1.0f, 1.0f, 1.0f );

		gi_mesh_draw( world->plane_mesh );

		glDisable( GL_BLEND );
	}

	glPopMatrix();

	//FIXME: don't use it here
	gi_ortho();

	/* cross */
	if ( world->ball.fpm )
	{
		glBindTexture( GL_TEXTURE_2D, 0 );
		glColor4f( 1.0f, 1.0f, 1.0f, 1.0f );

		glBegin( GL_LINES );

		glVertex2f( 0.5f + 0.01f, 0.5f + 0.01f );
		glVertex2f( 0.5f + 0.02f, 0.5f + 0.02f );

		glVertex2f( 0.5f - 0.01f, 0.5f + 0.01f );
		glVertex2f( 0.5f - 0.02f, 0.5f + 0.02f );

		glVertex2f( 0.5f + 0.01f, 0.5f - 0.01f );
		glVertex2f( 0.5f + 0.02f, 0.5f - 0.02f );

		glVertex2f( 0.5f - 0.01f, 0.5f - 0.01f );
		glVertex2f( 0.5f - 0.02f, 0.5f - 0.02f );

		glEnd();
	}

	/* text */
	if ( world->win )
	{
		glEnable( GL_BLEND );
		glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
		gi_string_draw_xy( world->gstr_win, 0.02f, 0.4f );
		glDisable( GL_BLEND );
	}
}

f32 g_world_get_h( g_world *world, f32 x, f32 y )
{
	int _x = ( int ) x;
	int _y = ( int ) y;
	f32 ex = x - ( f32 ) _x ;
	f32 ey = y - ( f32 ) _y ;

	if ( ( ex + ey ) < 1.0f )
	{
		f32 t0 = G_HEIGHT_SCALER * ( f32 )
			world->hmap.raw[ 3 * ( ( ( int ) y ) + ( ( int ) x ) * world->hmap.height ) ];
		f32 t1 = G_HEIGHT_SCALER * ( f32 )
			world->hmap.raw[ 3 * ( ( 1 + ( int ) y ) + ( ( int ) x ) * world->hmap.height ) ];
		f32 t2 = G_HEIGHT_SCALER * ( f32 )
			world->hmap.raw[ 3 * ( ( ( int ) y ) + ( 1 + ( int ) x ) * world->hmap.height ) ];
		return t0 + ( t1 - t0 ) * ey + ( t2 - t0 ) * ex;
	}
	else
	{
		f32 t1 = G_HEIGHT_SCALER * ( f32 )
			world->hmap.raw[ 3 * ( ( 1 + ( int ) y ) + ( ( int ) x ) * world->hmap.height ) ];
		f32 t2 = G_HEIGHT_SCALER * ( f32 )
			world->hmap.raw[ 3 * ( ( ( int ) y ) + ( 1 + ( int ) x ) * world->hmap.height ) ];
		f32 t3 = G_HEIGHT_SCALER * ( f32 )
			world->hmap.raw[ 3 * ( ( 1 + ( int ) y ) + ( 1 + ( int ) x ) * world->hmap.height ) ];
		return t3 + ( t1 - t3 ) * ( 1.0f - ex ) + ( t2 - t3 ) * ( 1.0f - ey );
	}
}

void g_world_get_point( g_world *world, f32 *beg, f32 *_ray, f32 *ret )
{
	f32 ray[ 3 ] = { _ray[ 0 ], _ray[ 1 ], _ray[ 2 ] };
	f32 ilen = ( f32 ) ( 1.0 / sqrt( ray[ 0 ] * ray[ 0 ] + ray[ 1 ] * ray[ 1 ] + ray[ 2 ] * ray[ 2 ] ) );
	ray[ 0 ] *= ilen;
	ray[ 1 ] *= ilen;
	ray[ 2 ] *= ilen;

	f32 dh ;
	f32 cur[ 3 ] = { beg[ 0 ], beg[ 1 ], beg[ 2 ] };

	int i = 0;
	do {
		cur[ 0 ] += ray[ 0 ];
		cur[ 1 ] += ray[ 1 ];
		cur[ 2 ] += ray[ 2 ];
		i++;

		f32 h = g_world_get_h( world, cur[ 0 ], cur[ 2 ] ); 
		dh = cur[ 1 ] - h;

		if ( cur[ 0 ] < 0.0f || cur[ 0 ] > world->hmap.width ||
				cur[ 2 ] < 0.0f || cur[ 2 ] > world->hmap.height )
			break;

	} while ( dh > 0.0f || i > 40 );

	ret[ 0 ] = cur[ 0 ];
	ret[ 1 ] = cur[ 1 ];
	ret[ 2 ] = cur[ 2 ];
}

