#define _XOPEN_SOURCE 500

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#define _GINT_INTERNAL_INCLUDE_
#include "gint.h"
#include "gimesh.h"

gi_mesh* gi_mesh_load( const char *file )
{
	gi_mesh *mesh = malloc( sizeof( gi_mesh ) );

	FILE *f = fopen( file, "rb" );

	if ( f == NULL )
	{
		fprintf( stderr, "%s:%i [ERROR] can't open file \"%s\" \n",
					__FILE__, __LINE__, file );
		return NULL;
	}

	fread( &mesh->icount, 2, 1, f );
	fread( &mesh->vcount, 2, 1, f );

	mesh->idata = malloc( mesh->icount * GI_MESH_ISIZE );
	fread( mesh->idata, GI_MESH_ISIZE * mesh->icount, 1, f );

	mesh->vdata = malloc( mesh->vcount * GI_MESH_VSIZE );
	fread( mesh->vdata, GI_MESH_VSIZE * mesh->vcount, 1, f );

	fclose( f );
	mesh->light = FALSE;
	
	return mesh;
}

gi_mesh* gi_mesh_create( 	ui16 vertex, ui16 index,
				void *v_data, void *i_data )
{
	gi_mesh *mesh = malloc( sizeof( gi_mesh ) );

	mesh->icount = index;
	mesh->vcount = vertex;

	mesh->idata = malloc( mesh->icount * GI_MESH_ISIZE );
	memcpy( mesh->idata, i_data, mesh->icount * GI_MESH_ISIZE );

	mesh->vdata = malloc( mesh->vcount * GI_MESH_VSIZE );
	memcpy( mesh->vdata, v_data, mesh->vcount * GI_MESH_VSIZE );

	mesh->light = FALSE;

	return mesh;
}

void gi_mesh_mklight( 	gi_mesh *mesh,
			const i8 *dir,
			ui8 ambient )
{
	int i ;
	for ( i = 0; i < mesh->vcount; i++ )
	{
		i8 n[ 4 ] = {
			*( i8* ) ( ( uint ) mesh->vdata + i * GI_MESH_VSIZE + 20 ),
			*( i8* ) ( ( uint ) mesh->vdata + i * GI_MESH_VSIZE + 21 ),
			*( i8* ) ( ( uint ) mesh->vdata + i * GI_MESH_VSIZE + 22 ),
	 		0 };
		int dot = ( n[ 0 ] * dir[ 0 ] + n[ 1 ] * dir[ 1 ] +
				n[ 2 ] * dir[ 2 ] ) / 64;
		if ( dot < 0 )
			dot = 0;
		dot += ambient;
		if ( dot > 255 )
			dot = 255;
		*( ui8* ) ( ( uint ) mesh->vdata + i * GI_MESH_VSIZE + 20 ) = ( ui8 ) dot;
		*( ui8* ) ( ( uint ) mesh->vdata + i * GI_MESH_VSIZE + 21 ) = ( ui8 ) dot;
		*( ui8* ) ( ( uint ) mesh->vdata + i * GI_MESH_VSIZE + 22 ) = ( ui8 ) dot;
		*( ui8* ) ( ( uint ) mesh->vdata + i * GI_MESH_VSIZE + 23 ) = 255;
	}

	mesh->light = TRUE;
}

void gi_mesh_draw( gi_mesh *mesh )
{
	glTexCoordPointer( 2, GL_FLOAT, GI_MESH_VSIZE,
			( void* ) ( ( uint ) mesh->vdata + 12 ) );
	if ( mesh->light )
		glColorPointer( 4, GL_UNSIGNED_BYTE, GI_MESH_VSIZE,
			( void* ) ( ( uint ) mesh->vdata + 20 ) );
	else
		glNormalPointer( GL_BYTE, GI_MESH_VSIZE,
			( void* ) ( ( uint ) mesh->vdata + 20 ) );
	glVertexPointer( 3, GL_FLOAT, GI_MESH_VSIZE, mesh->vdata );

	glEnableClientState( GL_TEXTURE_COORD_ARRAY );
	if ( mesh->light )
		glEnableClientState( GL_COLOR_ARRAY );
	else
		glEnableClientState( GL_NORMAL_ARRAY );
	glEnableClientState( GL_VERTEX_ARRAY );

	if ( mesh->idata != NULL )
		glDrawElements( GL_TRIANGLES, mesh->icount,
					GL_UNSIGNED_SHORT, mesh->idata );
	else
		glDrawArrays( GL_TRIANGLES, 0, mesh->vcount );

	glDisableClientState( GL_TEXTURE_COORD_ARRAY );
	if ( mesh->light )
		glDisableClientState( GL_COLOR_ARRAY );
	else
		glDisableClientState( GL_NORMAL_ARRAY );
	glDisableClientState( GL_VERTEX_ARRAY );
}

void gi_mesh_delete( gi_mesh *mesh )
{
	free( mesh->idata );
	free( mesh->vdata );
	free( mesh );
}

