#define _XOPEN_SOURCE 500

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#define _GINT_INTERNAL_INCLUDE_
#include "giui.h"
#include "gint.h"

enum {
	GI_UI_COLOR_BACKGROUND		= 0x0000,
	GI_UI_COLOR_FONT,
	GI_UI_COLOR_CURSOR,
	GI_UI_COLOR_UIE,
	GI_UI_COLOR_LIGHT,
	GI_UI_COLOR_PUSH,
	GI_UI_COLOR_ON,
	GI_UI_COLOR_OFF,
	GI_UI_COLOR_SELECT
};

gi_ui_base *gi_ui_base_create( gi_font *font )
{
	gi_ui_base *base = malloc( sizeof( gi_ui_base ) );

	base->font = font;

	/* theme */

	base->fdefsize = 0.025f;

	base->cltheme[ GI_UI_COLOR_BACKGROUND ][ 0 ] = 64;
	base->cltheme[ GI_UI_COLOR_BACKGROUND ][ 1 ] = 64;
	base->cltheme[ GI_UI_COLOR_BACKGROUND ][ 2 ] = 128;
	base->cltheme[ GI_UI_COLOR_BACKGROUND ][ 3 ] = 128;

	base->cltheme[ GI_UI_COLOR_FONT ][ 0 ] = 255;
	base->cltheme[ GI_UI_COLOR_FONT ][ 1 ] = 255;
	base->cltheme[ GI_UI_COLOR_FONT ][ 2 ] = 255;
	base->cltheme[ GI_UI_COLOR_FONT ][ 3 ] = 196;

	base->cltheme[ GI_UI_COLOR_CURSOR ][ 0 ] = 196;
	base->cltheme[ GI_UI_COLOR_CURSOR ][ 1 ] = 196;
	base->cltheme[ GI_UI_COLOR_CURSOR ][ 2 ] = 255;
	base->cltheme[ GI_UI_COLOR_CURSOR ][ 3 ] = 128;

	base->cltheme[ GI_UI_COLOR_UIE ][ 0 ] = 64;
	base->cltheme[ GI_UI_COLOR_UIE ][ 1 ] = 64;
	base->cltheme[ GI_UI_COLOR_UIE ][ 2 ] = 128;
	base->cltheme[ GI_UI_COLOR_UIE ][ 3 ] = 128;

	base->cltheme[ GI_UI_COLOR_LIGHT ][ 0 ] = 96;
	base->cltheme[ GI_UI_COLOR_LIGHT ][ 1 ] = 96;
	base->cltheme[ GI_UI_COLOR_LIGHT ][ 2 ] = 96;
	base->cltheme[ GI_UI_COLOR_LIGHT ][ 3 ] = 64;

	base->cltheme[ GI_UI_COLOR_PUSH ][ 0 ] = 32;
	base->cltheme[ GI_UI_COLOR_PUSH ][ 1 ] = 32;
	base->cltheme[ GI_UI_COLOR_PUSH ][ 2 ] = 32;
	base->cltheme[ GI_UI_COLOR_PUSH ][ 3 ] = 64;

	base->cltheme[ GI_UI_COLOR_ON ][ 0 ] = 128;
	base->cltheme[ GI_UI_COLOR_ON ][ 1 ] = 128;
	base->cltheme[ GI_UI_COLOR_ON ][ 2 ] = 128;
	base->cltheme[ GI_UI_COLOR_ON ][ 3 ] = 64;

	base->cltheme[ GI_UI_COLOR_OFF ][ 0 ] = 16;
	base->cltheme[ GI_UI_COLOR_OFF ][ 1 ] = 16;
	base->cltheme[ GI_UI_COLOR_OFF ][ 2 ] = 16;
	base->cltheme[ GI_UI_COLOR_OFF ][ 3 ] = 64;

	base->cltheme[ GI_UI_COLOR_SELECT ][ 0 ] = 64;
	base->cltheme[ GI_UI_COLOR_SELECT ][ 1 ] = 64;
	base->cltheme[ GI_UI_COLOR_SELECT ][ 2 ] = 255;
	base->cltheme[ GI_UI_COLOR_SELECT ][ 3 ] = 64;

	/* end of theme */

	base->vupt = malloc( 3 * ( 8 ) );
	* ( f32* ) ( ( uint ) base->vupt + 8 * 0 + 0 ) = 0.0f;
	* ( f32* ) ( ( uint ) base->vupt + 8 * 0 + 4 ) = 0.0f;
	* ( f32* ) ( ( uint ) base->vupt + 8 * 1 + 0 ) = base->fdefsize * 0.4f;
	* ( f32* ) ( ( uint ) base->vupt + 8 * 1 + 4 ) = base->fdefsize;
	* ( f32* ) ( ( uint ) base->vupt + 8 * 2 + 0 ) = base->fdefsize;
	* ( f32* ) ( ( uint ) base->vupt + 8 * 2 + 4 ) = base->fdefsize * 0.5f;

	base->vsel = malloc( 5 * ( 8 ) );
	* ( f32* ) ( ( uint ) base->vsel + 8 * 0 + 0 ) = 0.0f;
	* ( f32* ) ( ( uint ) base->vsel + 8 * 0 + 4 ) = 0.0f;
	* ( f32* ) ( ( uint ) base->vsel + 8 * 1 + 0 ) = 0.0f;
	* ( f32* ) ( ( uint ) base->vsel + 8 * 1 + 4 ) = 1.0f;
	* ( f32* ) ( ( uint ) base->vsel + 8 * 2 + 0 ) = 1.0f;
	* ( f32* ) ( ( uint ) base->vsel + 8 * 2 + 4 ) = 1.0f;
	* ( f32* ) ( ( uint ) base->vsel + 8 * 3 + 0 ) = 1.0f;
	* ( f32* ) ( ( uint ) base->vsel + 8 * 3 + 4 ) = 0.0f;
	* ( f32* ) ( ( uint ) base->vsel + 8 * 4 + 0 ) = 0.0f;
	* ( f32* ) ( ( uint ) base->vsel + 8 * 4 + 4 ) = 0.0f;

	/* (4) - maximum of flying UIEs */
	base->fly_uie_list = malloc( 4 * sizeof( gi_uie_fly ) );
	memset( base->fly_uie_list, 0, 4 * sizeof( gi_uie_fly ) );

	/* (8) - maximum of root UIEs count */
	base->root_uie_list = malloc( 8 * sizeof( gi_uie* ) );
	memset( base->root_uie_list, 0, 8 * sizeof( gi_uie* ) );

	int iw, ih;
	gi_get( GI_WINDOW_WIDTH, &iw );
	gi_get( GI_WINDOW_HEIGHT, &ih );

	base->wnd_w = ( f32 ) iw;
	base->wnd_h = ( f32 ) ih;

	base->l_uie = NULL;
	base->s_uie = NULL;
	base->upt_on = TRUE;
	base->move = FALSE;
	base->fly = FALSE;
	base->lref = FALSE;
	base->glref = TRUE;

	return base;
}

static void gi_uie_delete_no_unreg( gi_uie *uie )
{
	uint i = 0;
	if ( uie->next_list != NULL )
	{
		while ( uie->next_list[ i ] != NULL )
			gi_uie_delete_no_unreg( uie->next_list[ i++ ] );
		free( uie->next_list );
	}

	switch ( uie->type )
	{
		case GI_UIE_WINDOW:
		case GI_UIE_BUTTON:
		case GI_UIE_TEXT:
		case GI_UIE_EDIT:
		case GI_UIE_CHECK:
			gi_string_delete( uie->text.string );
			break;

		case GI_UIE_COMBO:
		case GI_UIE_LIST:
			i = 0;
			while ( uie->text.list[ i ] != NULL )
				gi_string_delete( uie->text.list[ i++ ] );
			free( uie->text.list );
			break;

		case GI_UIE_IMAGE:
			break;
	}

	if ( uie->info != NULL )
		free( uie->info );
	if ( uie->vdata != NULL )
		free( uie->vdata );
	free( uie );
}

void gi_ui_base_delete( gi_ui_base *base )
{
	int i ;
	for ( i = 0; i < 8; i++ )
	{
		if ( base->root_uie_list[ i ] != NULL )
			gi_uie_delete_no_unreg( GIUI_UIE( base->root_uie_list[ i ] ) );
	}

	free( base->vupt );
	free( base->vsel );
	free( base->fly_uie_list );
	free( base->root_uie_list );
	free( base );
}

static gi_uie *gi_uie_find( gi_uie *uie, f32 x, f32 y )
{
	gi_uie *ret = NULL;

	if ( ( uie->x < x ) && ( ( uie->x + uie->w ) > x ) &&
			( uie->y < y ) && ( ( uie->y + uie->h ) > y ) )
	{
		ret = uie;
		if ( uie->next_list != NULL )
		{
			gi_uie *next_ret = NULL;
			uint i = 0;
			while ( ( uie->next_list[ i ] != NULL ) &&
					( next_ret == NULL ) )
				next_ret = gi_uie_find( uie->next_list[ i++ ],
						x - uie->x, y - uie->y );
			if ( next_ret != NULL )
				ret = next_ret;
		}
	}

	return ret;
}

static void gi_get_uie_absxy( gi_uie *uie, f32 *x, f32 *y )
{
	gi_uie *c_uie = uie;
	*x = *y = 0.0f;

	while ( c_uie->last != NULL )
	{
		*x += c_uie->x;
		*y += c_uie->y;
		c_uie = c_uie->last;
	}

	uint i = 0;
	while ( GIUI_UIE( c_uie->base->root_uie_list[ i ] ) != c_uie ) i++;
	*x += GIUI_UIE( c_uie->base->root_uie_list[ i ] )->x;
	*y += GIUI_UIE( c_uie->base->root_uie_list[ i ] )->y;
}

void gi_ui_update( gi_ui_base *base, gi_input *input, f64 dt )
{
	base->upt_x = ( ( f32 ) input->mouse_x ) / base->wnd_w;
	base->upt_y = ( ( f32 ) input->mouse_y ) / base->wnd_h;
	gi_uie *l_uie = GIUI_UIE( base->l_uie );
	gi_uie *s_uie = GIUI_UIE( base->s_uie );
	base->glref = FALSE;

	if ( l_uie != NULL )
	{
		if ( l_uie->flags & GI_UIE_MOVEABLE )
		{
			if ( ( input->keys[ GI_BUTTON_LEFT ] == GI_KEY_DOWN ) &&
					( base->move == FALSE ) )
			{
				base->move = TRUE;
				base->move_upt_x = base->upt_x;
				base->move_upt_y = base->upt_y;
				base->move_uie_x = l_uie->x;
				base->move_uie_y = l_uie->y;
				if ( l_uie->func != NULL )
					l_uie->func( l_uie, GI_UIE_MOVE );
				base->glref = TRUE;
				input->keys[ GI_BUTTON_LEFT ] = GI_KEY_OFF;
			}

			if ( ( input->keys[ GI_BUTTON_LEFT ] == GI_KEY_UP ) &&
					( base->move ) )
			{
				base->move = FALSE;
				if ( l_uie->func != NULL )
					l_uie->func( l_uie, GI_UIE_UNMOVE );
				base->glref = TRUE;
				input->keys[ GI_BUTTON_LEFT ] = GI_KEY_OFF;
			}

			if ( base->move )
			{
				l_uie->x = base->move_uie_x + 
					( base->upt_x - base->move_upt_x );
				l_uie->y = base->move_uie_y + 
					( base->upt_y - base->move_upt_y );
				base->glref = TRUE;
			}
		}

		if ( ( l_uie->flags & GI_UIE_BORDERED ) && ( base->move ) )
		{
			if ( l_uie->x < 0.0f )
				l_uie->x = 0.0f;
			if ( ( l_uie->x + l_uie->w ) > 1.0f )
				l_uie->x = 1.0f - l_uie->w;
			if ( l_uie->y < 0.0f )
				l_uie->y = 0.0f;
			if ( ( l_uie->y + l_uie->h ) > 1.0f )
				l_uie->y = 1.0f - l_uie->h;
		}

		if ( ( l_uie->next_list != NULL ) ||
				( ( l_uie->x > base->upt_x ) ||
				  ( ( l_uie->x + l_uie->w ) < base->upt_x ) ||
				  ( l_uie->y > base->upt_y ) ||
				  ( ( l_uie->y + l_uie->h ) < base->upt_y ) ) )
		{
			l_uie = NULL;
			uint i = 0 ;
			while ( ( base->root_uie_list[ i ] != NULL ) &&
					( l_uie == NULL ) )
				l_uie = gi_uie_find( base->root_uie_list[ i++ ],
						base->upt_x, base->upt_y );
		}
	}
	else
	{
		l_uie = NULL;
		uint i = 0 ;
		while ( ( base->root_uie_list[ i ] != NULL ) &&
				( l_uie == NULL ) )
			l_uie = gi_uie_find( GIUI_UIE( base->root_uie_list[ i++ ] ),
					base->upt_x, base->upt_y );
	}

	if ( ( base->l_uie != l_uie ) && ( l_uie != NULL ) )
		if ( ( l_uie->flags & GI_UIE_NOSELECT ) == 0 )
			base->s_uie = l_uie;

	if ( ( base->l_uie != l_uie ) || base->lref )
	{
		if ( base->l_uie != NULL )
		{
			switch ( GIUI_UIE( base->l_uie )->type )
			{
				case GI_UIE_BUTTON:
					GIUI_BUTTON_INFO( ( gi_uie* ) base->l_uie )->light = FALSE;
					GIUI_BUTTON_INFO( ( gi_uie* ) base->l_uie )->push = FALSE;
					break;

				case GI_UIE_CHECK:
					GIUI_CHECK_INFO( ( gi_uie* ) base->l_uie )->light = FALSE;
					GIUI_CHECK_INFO( ( gi_uie* ) base->l_uie )->push = FALSE;
					break;

				case GI_UIE_COMBO:
					GIUI_COMBO_INFO( ( gi_uie* ) base->l_uie )->light = FALSE;
					if ( base->l_uie != l_uie ) {
						GIUI_COMBO_INFO( ( gi_uie* ) base->l_uie )->l_push = FALSE;
						GIUI_COMBO_INFO( ( gi_uie* ) base->l_uie )->r_push = FALSE;
					}
					break;

				case GI_UIE_LIST:
					if ( base->l_uie != l_uie )
						GIUI_LIST_INFO( ( gi_uie* ) base->l_uie )->light = 0;
					break;
			}
		}

		if ( l_uie != NULL )
		{
			uint tmp ;
			f32 absx, absy ;
			switch ( l_uie->type )
			{
				case GI_UIE_BUTTON:
					GIUI_BUTTON_INFO( l_uie )->light = TRUE;
					break;

				case GI_UIE_CHECK:
					GIUI_CHECK_INFO( l_uie )->light = TRUE;
					break;

				case GI_UIE_COMBO:
					GIUI_COMBO_INFO( l_uie )->light = TRUE;
					break;

				case GI_UIE_LIST:
					gi_get_uie_absxy( l_uie, &absx, &absy );
					tmp = GIUI_LIST_INFO( l_uie )->light;
					GIUI_LIST_INFO( l_uie )->light = 1 + ( uint ) 
						( ( base->upt_y - absy ) /
						  ( base->fdefsize * 1.5f ) ) +
						GIUI_LIST_INFO( l_uie )->scroll;
					if ( tmp != GIUI_LIST_INFO( l_uie )->light )
						base->glref = TRUE;
					break;
			}

			base->lref = ( l_uie->type == GI_UIE_COMBO ) ||
				( l_uie->type == GI_UIE_LIST ) ;
		}

		if ( base->l_uie != l_uie ) base->glref = TRUE;
		base->l_uie = l_uie ;
	}

	if ( s_uie != NULL )
	{
		gi_uie *top = s_uie->last;
		gi_uie **nlist;
		if ( top != NULL )
			nlist = top->next_list;
		else
			nlist = base->root_uie_list;

		if ( input->keys[ GI_TAB ] == GI_KEY_DOWN ||
				input->keys[ GI_N ] == GI_KEY_DOWN )
		{
			int i = 0;
			while ( nlist[ i ] != NULL )
				if ( nlist[ i ] == s_uie )
				{
					do {
						s_uie = ( nlist[ i + 1 ] != NULL ) ?
							nlist[ i + 1 ] : nlist[ 0 ] ;
						i++;
						if ( nlist[ i ] == NULL )
							i = 0;
					}
					while ( s_uie->flags & GI_UIE_NOSELECT );
					base->s_uie = s_uie;
					break;
				}
				else
					i++;

			input->keys[ GI_TAB ] = GI_KEY_OFF;
			input->keys[ GI_N ] = GI_KEY_OFF;
			base->glref = TRUE;
		}

		if ( input->keys[ GI_P ] == GI_KEY_DOWN )
		{
			int i = 0;
			while ( nlist[ i ] != NULL )
				if ( nlist[ i ] == s_uie )
				{
					int len = 0;
					while ( nlist[ len ] != NULL ) len++;
					do {
						s_uie = ( ( i - 1 ) >= 0 ) ?
							nlist[ i - 1 ] : nlist[ len - 1 ] ;
						i--;
						if ( i < 0 )
							i = len - 1;
					}
					while ( s_uie->flags & GI_UIE_NOSELECT );
					base->s_uie = s_uie;
					break;
				}
				else
					i++;

			input->keys[ GI_P ] = GI_KEY_OFF;
			base->glref = TRUE;
		}

		if ( s_uie->flags & GI_UIE_PUSHABLE )
		{
			switch ( s_uie->type )
			{
				f32 absx, absy ;
				case GI_UIE_BUTTON:
				if ( ( input->keys[ GI_BUTTON_LEFT ] == GI_KEY_DOWN ||
							input->keys[ GI_ENTER ] == GI_KEY_DOWN ||
							input->keys[ GI_SPACE ] == GI_KEY_DOWN ) &&
						( GIUI_BUTTON_INFO( s_uie )->push == FALSE ) )
				{
					GIUI_BUTTON_INFO( s_uie )->push = TRUE;
					if ( s_uie->func != NULL )
						s_uie->func( s_uie, GI_UIE_PUSH );
					base->glref = TRUE;
					input->keys[ GI_BUTTON_LEFT ] = GI_KEY_OFF;
					input->keys[ GI_ENTER ] = GI_KEY_OFF;
					input->keys[ GI_SPACE ] = GI_KEY_OFF;
				}
				if ( ( input->keys[ GI_BUTTON_LEFT ] == GI_KEY_UP ||
							input->keys[ GI_ENTER ] == GI_KEY_UP ||
							input->keys[ GI_SPACE ] == GI_KEY_UP ) &&
						( GIUI_BUTTON_INFO( s_uie )->push ) )
				{
					GIUI_BUTTON_INFO( s_uie )->push = FALSE;
					if ( s_uie->func != NULL )
						s_uie->func( s_uie, GI_UIE_UNPUSH );
					base->glref = TRUE;
					input->keys[ GI_BUTTON_LEFT ] = GI_KEY_OFF;
					input->keys[ GI_ENTER ] = GI_KEY_OFF;
					input->keys[ GI_SPACE ] = GI_KEY_OFF;
				}
				break;

				case GI_UIE_CHECK:
				if ( ( input->keys[ GI_BUTTON_LEFT ] == GI_KEY_DOWN ||
							input->keys[ GI_ENTER ] == GI_KEY_DOWN ||
							input->keys[ GI_SPACE ] == GI_KEY_DOWN ) &&
						( GIUI_CHECK_INFO( s_uie )->push == FALSE ) )
				{
					GIUI_CHECK_INFO( s_uie )->push = TRUE;
					GIUI_CHECK_INFO( s_uie )->check =
						!GIUI_CHECK_INFO( s_uie )->check ;
					if ( s_uie->func != NULL )
						s_uie->func( s_uie, GI_UIE_PUSH );
					base->glref = TRUE;
					input->keys[ GI_BUTTON_LEFT ] = GI_KEY_OFF;
					input->keys[ GI_ENTER ] = GI_KEY_OFF;
					input->keys[ GI_SPACE ] = GI_KEY_OFF;
				}
				if ( ( input->keys[ GI_BUTTON_LEFT ] == GI_KEY_UP ||
							input->keys[ GI_ENTER ] == GI_KEY_UP ||
							input->keys[ GI_SPACE ] == GI_KEY_UP ) &&
						( GIUI_CHECK_INFO( s_uie )->push ) )
				{
					GIUI_CHECK_INFO( s_uie )->push = FALSE;
					if ( s_uie->func != NULL )
						s_uie->func( s_uie, GI_UIE_UNPUSH );
					base->glref = TRUE;
					input->keys[ GI_BUTTON_LEFT ] = GI_KEY_OFF;
					input->keys[ GI_ENTER ] = GI_KEY_OFF;
					input->keys[ GI_SPACE ] = GI_KEY_OFF;
				}
				break;

				case GI_UIE_COMBO:
				if ( ( input->keys[ GI_BUTTON_LEFT ] == GI_KEY_DOWN ||
							input->keys[ GI_H ] == GI_KEY_DOWN ||
							input->keys[ GI_L ] == GI_KEY_DOWN ||
							input->keys[ GI_LEFT ] == GI_KEY_DOWN ||
							input->keys[ GI_RIGHT ] == GI_KEY_DOWN ) &&
						( ( GIUI_COMBO_INFO( s_uie )->l_push == FALSE ) &&
						  ( GIUI_COMBO_INFO( s_uie )->r_push == FALSE ) ) )
				{
					gi_get_uie_absxy( s_uie, &absx, &absy );
					if ( ( fabsf( absx - base->upt_x ) < base->fdefsize * 1.5f ) ||
							( input->keys[ GI_H ] == GI_KEY_DOWN ||
							  input->keys[ GI_LEFT ] == GI_KEY_DOWN ) )
					{
						GIUI_COMBO_INFO( s_uie )->l_push = TRUE;
						if ( GIUI_COMBO_INFO( s_uie )->index > 1 )
							GIUI_COMBO_INFO( s_uie )->index--;
						if ( s_uie->func != NULL )
							s_uie->func( s_uie, GI_UIE_PUSH );
						base->glref = TRUE;
					}
					absx += s_uie->w;
					if ( ( fabsf( absx - base->upt_x ) < base->fdefsize * 1.5f ) ||
							( input->keys[ GI_L ] == GI_KEY_DOWN ||
							  input->keys[ GI_RIGHT ] == GI_KEY_DOWN ) )
					{
						GIUI_COMBO_INFO( s_uie )->r_push = TRUE;
						if ( GIUI_COMBO_INFO( s_uie )->index < 
								GIUI_COMBO_INFO( s_uie )->length )
							GIUI_COMBO_INFO( s_uie )->index++;
						if ( s_uie->func != NULL )
							s_uie->func( s_uie, GI_UIE_PUSH );
						base->glref = TRUE;
					}
					input->keys[ GI_BUTTON_LEFT ] = GI_KEY_OFF;
					input->keys[ GI_H ] = GI_KEY_OFF;
					input->keys[ GI_L ] = GI_KEY_OFF;
					input->keys[ GI_LEFT ] = GI_KEY_OFF;
					input->keys[ GI_RIGHT ] = GI_KEY_OFF;
				}
				if ( ( input->keys[ GI_BUTTON_LEFT ] == GI_KEY_UP ||
							input->keys[ GI_H ] == GI_KEY_UP ||
							input->keys[ GI_L ] == GI_KEY_UP ||
							input->keys[ GI_LEFT ] == GI_KEY_UP ||
							input->keys[ GI_RIGHT ] == GI_KEY_UP ) &&
						( ( GIUI_COMBO_INFO( s_uie )->l_push ) ||
						  ( GIUI_COMBO_INFO( s_uie )->r_push ) ) )
				{
					gi_get_uie_absxy( s_uie, &absx, &absy );
					if ( ( fabsf( absx - base->upt_x ) < base->fdefsize * 1.5f ) ||
							( input->keys[ GI_H ] == GI_KEY_UP ||
							  input->keys[ GI_LEFT ] == GI_KEY_UP ) )
					{
						GIUI_COMBO_INFO( s_uie )->l_push = FALSE;
						if ( s_uie->func != NULL )
							s_uie->func( s_uie, GI_UIE_UNPUSH );
						base->glref = TRUE;
					}
					absx += s_uie->w;
					if ( ( fabsf( absx - base->upt_x ) < base->fdefsize * 1.5f ) ||
							( input->keys[ GI_L ] == GI_KEY_UP ||
							  input->keys[ GI_RIGHT ] == GI_KEY_UP ) )
					{
						GIUI_COMBO_INFO( s_uie )->r_push = FALSE;
						if ( s_uie->func != NULL )
							s_uie->func( s_uie, GI_UIE_UNPUSH );
						base->glref = TRUE;
					}
					input->keys[ GI_BUTTON_LEFT ] = GI_KEY_OFF;
					input->keys[ GI_H ] = GI_KEY_OFF;
					input->keys[ GI_L ] = GI_KEY_OFF;
					input->keys[ GI_LEFT ] = GI_KEY_OFF;
					input->keys[ GI_RIGHT ] = GI_KEY_OFF;
				}
				break;

				case GI_UIE_LIST:
				if ( ( input->keys[ GI_BUTTON_LEFT ] == GI_KEY_DOWN ) &&
						( GIUI_LIST_INFO( s_uie )->push == FALSE ) )
				{
					GIUI_LIST_INFO( s_uie )->push = TRUE;
					gi_get_uie_absxy( s_uie, &absx, &absy );
					GIUI_LIST_INFO( s_uie )->index = 1 + ( uint ) 
						( ( base->upt_y - absy ) /
						  ( base->fdefsize * 1.5f ) ) +
						GIUI_LIST_INFO( s_uie )->scroll;
					if ( s_uie->func != NULL )
						s_uie->func( s_uie, GI_UIE_PUSH );
					base->glref = TRUE;
					input->keys[ GI_BUTTON_LEFT ] = GI_KEY_OFF;
				}
				if ( input->keys[ GI_J ] == GI_KEY_DOWN ||
							input->keys[ GI_DOWN ] == GI_KEY_DOWN )
				{
					if ( GIUI_LIST_INFO( s_uie )->index <
							GIUI_LIST_INFO( s_uie )->length )
						GIUI_LIST_INFO( s_uie )->index += 1;
					if ( GIUI_LIST_INFO( s_uie )->index >
							GIUI_LIST_INFO( s_uie )->size +
							GIUI_LIST_INFO( s_uie )->scroll )
						if ( GIUI_LIST_INFO( s_uie )->scroll < 
								( GIUI_LIST_INFO( s_uie )->length -
								  GIUI_LIST_INFO( s_uie )->size ) )
							GIUI_LIST_INFO( s_uie )->scroll++;
					if ( s_uie->func != NULL ) {
						s_uie->func( s_uie, GI_UIE_PUSH );
						s_uie->func( s_uie, GI_UIE_UNPUSH );
					}
					base->glref = TRUE;
					input->keys[ GI_J ] = GI_KEY_OFF;
					input->keys[ GI_DOWN ] = GI_KEY_OFF;
				}
				if ( input->keys[ GI_K ] == GI_KEY_DOWN ||
							input->keys[ GI_UP ] == GI_KEY_DOWN )
				{
					if ( GIUI_LIST_INFO( s_uie )->index > 1 )
						GIUI_LIST_INFO( s_uie )->index -= 1;
					if ( GIUI_LIST_INFO( s_uie )->index - 1 <
							GIUI_LIST_INFO( s_uie )->scroll )
						if ( GIUI_LIST_INFO( s_uie )->scroll > 0 )
							GIUI_LIST_INFO( s_uie )->scroll--;
					if ( s_uie->func != NULL ) {
						s_uie->func( s_uie, GI_UIE_PUSH );
						s_uie->func( s_uie, GI_UIE_UNPUSH );
					}
					base->glref = TRUE;
					input->keys[ GI_K ] = GI_KEY_OFF;
					input->keys[ GI_UP ] = GI_KEY_OFF;
				}
				if ( ( input->keys[ GI_BUTTON_LEFT ] == GI_KEY_UP ) &&
						( GIUI_LIST_INFO( s_uie )->push ) )
				{
					GIUI_LIST_INFO( s_uie )->push = FALSE;
					if ( s_uie->func != NULL )
						s_uie->func( s_uie, GI_UIE_UNPUSH );
					base->glref = TRUE;
					input->keys[ GI_BUTTON_LEFT ] = GI_KEY_OFF;
				}
				if ( input->keys[ GI_BUTTON_SCROLL_UP ] == GI_KEY_DOWN )
				{
					if ( GIUI_LIST_INFO( s_uie )->scroll > 0 )
						GIUI_LIST_INFO( s_uie )->scroll--;
					base->glref = TRUE;
					input->keys[ GI_BUTTON_SCROLL_UP ] = GI_KEY_OFF;
				}
				if ( input->keys[ GI_BUTTON_SCROLL_DOWN ] == GI_KEY_DOWN )
				{
					if ( GIUI_LIST_INFO( s_uie )->scroll < 
							( GIUI_LIST_INFO( s_uie )->length -
							  GIUI_LIST_INFO( s_uie )->size ) )
						GIUI_LIST_INFO( s_uie )->scroll++;
					base->glref = TRUE;
					input->keys[ GI_BUTTON_SCROLL_DOWN ] = GI_KEY_OFF;
				}
				break;
			}
		}
	}

	gi_uie_fly *list = base->fly_uie_list;

	if ( base->fly )
	{
		base->fly = FALSE;
		base->glref = TRUE;

		uint i ;
		for ( i = 0; i < 4; i++ )
		{
			if ( list[ i ].use )
			{
				base->fly = TRUE;

				if ( list[ i ].time > 0.0f )
				{
					list[ i ].pos = list[ i ].acc * list[ i ].time *
						list[ i ].time * 0.5f;
					list[ i ].target->x = list[ i ].dest_x +
						( list[ i ].src_x - list[ i ].dest_x ) * list[ i ].pos;
					list[ i ].target->y = list[ i ].dest_y +
						( list[ i ].src_y - list[ i ].dest_y ) * list[ i ].pos;
					list[ i ].time -= dt;
				}
				else
				{
					list[ i ].target->x = list[ i ].dest_x;
					list[ i ].target->y = list[ i ].dest_y;
					list[ i ].use = FALSE;
				}
			}
		}
	}
}

void gi_uie_draw( gi_uie *uie )
{
	if ( ( uie->flags & GI_UIE_HIDDEN ) == 0 );
	{
		glBindTexture( GL_TEXTURE_2D, 0 );
		glPushMatrix();

		glTranslatef( uie->x, uie->y, 0.0f );

		uint i ;
		switch ( uie->type )
		{
			case GI_UIE_WINDOW:
				glColor4ubv( uie->base->cltheme[ GI_UI_COLOR_BACKGROUND ] );
				glVertexPointer( 2, GL_FLOAT, 8, uie->vdata );
				glEnableClientState( GL_VERTEX_ARRAY );
				glDrawArrays( GL_QUADS, 0, 8 );
				glDisableClientState( GL_VERTEX_ARRAY );
				gi_string_draw_xy( uie->text.string, uie->tx, uie->ty );
				break;

			case GI_UIE_BUTTON:
				glColor4ubv( uie->base->cltheme[ GI_UI_COLOR_UIE ] );
				glVertexPointer( 2, GL_FLOAT, 8, uie->vdata );
				glEnableClientState( GL_VERTEX_ARRAY );
				glDrawArrays( GL_QUADS, 0, 4 );
				if ( GIUI_BUTTON_INFO( uie )->push ) {
					glColor4ubv( uie->base->cltheme[ GI_UI_COLOR_PUSH ] );
					glDrawArrays( GL_QUADS, 0, 4 );
				}
				else
					if ( GIUI_BUTTON_INFO( uie )->light )	{
						glColor4ubv( uie->base->cltheme[ GI_UI_COLOR_LIGHT ] );
						glDrawArrays( GL_QUADS, 0, 4 );
					}
				glDisableClientState( GL_VERTEX_ARRAY );
				gi_string_draw_xy( uie->text.string, uie->tx, uie->ty );
				break;

			case GI_UIE_TEXT:
				gi_string_draw_xy( uie->text.string, uie->tx, uie->ty );
				break;

			case GI_UIE_CHECK:
				glColor4ubv( uie->base->cltheme[ GI_UI_COLOR_OFF ] );
				glVertexPointer( 2, GL_FLOAT, 8, uie->vdata );
				glEnableClientState( GL_VERTEX_ARRAY );
				if ( GIUI_CHECK_INFO( uie )->check ) {
					glColor4ubv( uie->base->cltheme[ GI_UI_COLOR_ON ] );
					glDrawArrays( GL_QUADS, 4, 4 );
				}
				else {
					glColor4ubv( uie->base->cltheme[ GI_UI_COLOR_OFF ] );
					glDrawArrays( GL_QUADS, 4, 4 );
				}
				if ( GIUI_CHECK_INFO( uie )->light )	{
					glColor4ubv( uie->base->cltheme[ GI_UI_COLOR_LIGHT ] );
					glDrawArrays( GL_QUADS, 0, 4 );
				}
				glDisableClientState( GL_VERTEX_ARRAY );
				gi_string_draw_xy( uie->text.string, uie->tx, uie->ty );
				break;

			case GI_UIE_COMBO:
				glColor4ubv( uie->base->cltheme[ GI_UI_COLOR_UIE ] );
				glVertexPointer( 2, GL_FLOAT, 8, uie->vdata );
				glEnableClientState( GL_VERTEX_ARRAY );
				glDrawArrays( GL_TRIANGLES, 4, 6 );
				if ( GIUI_COMBO_INFO( uie )->l_push )	{
					glColor4ubv( uie->base->cltheme[ GI_UI_COLOR_ON ] );
					glDrawArrays( GL_TRIANGLES, 4, 3 );
				}
				else
					if ( GIUI_COMBO_INFO( uie )->r_push )	{
						glColor4ubv( uie->base->cltheme[ GI_UI_COLOR_ON ] );
						glDrawArrays( GL_TRIANGLES, 7, 3 );
					}
				if ( GIUI_COMBO_INFO( uie )->light )	{
					glColor4ubv( uie->base->cltheme[ GI_UI_COLOR_LIGHT ] );
					glDrawArrays( GL_QUADS, 0, 4 );
				}
				glDisableClientState( GL_VERTEX_ARRAY );
				gi_string_draw_xy( uie->text.list[ 
						GIUI_COMBO_INFO( uie )->index - 1 ],
						uie->tx, uie->ty );
				break;

			case GI_UIE_LIST:
				glColor4ubv( uie->base->cltheme[ GI_UI_COLOR_UIE ] );
				glVertexPointer( 2, GL_FLOAT, 8, uie->vdata );
				glEnableClientState( GL_VERTEX_ARRAY );
				glDrawArrays( GL_QUADS, 0, 4 );
				if ( ( GIUI_LIST_INFO( uie )->index > 
							GIUI_LIST_INFO( uie )->scroll ) &&
						( GIUI_LIST_INFO( uie )->index < 
						  ( GIUI_LIST_INFO( uie )->scroll +
						    GIUI_LIST_INFO( uie )->size + 1 ) ) )
				{
					glPushMatrix();
					glTranslatef( 0.0f, ( GIUI_LIST_INFO( uie )->index -
								GIUI_LIST_INFO( uie )->scroll - 1 ) *
							uie->base->fdefsize * 1.5f, 0.0f );
					glColor4ubv( uie->base->cltheme[ GI_UI_COLOR_SELECT ] );
					glDrawArrays( GL_QUADS, 4, 4 );
					glPopMatrix();
				}
				if ( ( GIUI_LIST_INFO( uie )->light > 
							GIUI_LIST_INFO( uie )->scroll ) &&
						( GIUI_LIST_INFO( uie )->light < 
						  ( GIUI_LIST_INFO( uie )->scroll +
						    GIUI_LIST_INFO( uie )->size + 1 ) ) )
				{
					glPushMatrix();
					glTranslatef( 0.0f, ( GIUI_LIST_INFO( uie )->light -
								GIUI_LIST_INFO( uie )->scroll - 1 ) *
							uie->base->fdefsize * 1.5f, 0.0f );
					glColor4ubv( uie->base->cltheme[ GI_UI_COLOR_LIGHT ] );
					glDrawArrays( GL_QUADS, 4, 4 );
					glPopMatrix();
				}
				glDisableClientState( GL_VERTEX_ARRAY );
				for ( i = 0; i < GIUI_LIST_INFO( uie )->size; i++ )
					gi_string_draw_xy( uie->text.list[ i +
							GIUI_LIST_INFO( uie )->scroll ],
							uie->tx, uie->ty +
							1.5f * uie->base->fdefsize * ( f32 ) i );
				break;
		}

		glBindTexture( GL_TEXTURE_2D, 0 );

		if ( uie == uie->base->s_uie )
		{
			glPushMatrix();
			glScalef( uie->w, uie->h, 1.0f );

			glColor4ubv( uie->base->cltheme[ GI_UI_COLOR_CURSOR ] );
			glVertexPointer( 2, GL_FLOAT, 8, uie->base->vsel );
			glEnableClientState( GL_VERTEX_ARRAY );
			glDrawArrays( GL_LINE_STRIP, 0, 5 );
			glDisableClientState( GL_VERTEX_ARRAY );
			glPopMatrix();
		}

		if ( uie->next_list != NULL )
		{
			uint i = 0;
			while ( uie->next_list[ i ] != NULL )
				gi_uie_draw( uie->next_list[ i++ ] );
		}

		glPopMatrix();
	}
}

void gi_ui_draw( gi_ui_base *base )
{
	glEnable( GL_BLEND );
	glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );

	glLoadIdentity();

	uint i = 0;
	while ( base->root_uie_list[ i ] != NULL )
	{
		gi_uie *uie = GIUI_UIE( base->root_uie_list[ i ] ) ;
		if ( ( ( uie->x + uie->w ) > 0.0f ) &&
				( ( uie->y + uie->h ) > 0.0f ) &&
				( ( uie->x ) < 1.0f ) && ( ( uie->x ) < 1.0f ) )
			gi_uie_draw( GIUI_UIE( base->root_uie_list[ i ] ) );
		i++;
	}

	if ( base->upt_on )
	{
		glBindTexture( GL_TEXTURE_2D, 0 );

		glLoadIdentity();
		glTranslatef( base->upt_x, base->upt_y, 0.0f );

		glColor4ubv( base->cltheme[ GI_UI_COLOR_CURSOR ] );
		glVertexPointer( 2, GL_FLOAT, 8, base->vupt );
		glEnableClientState( GL_VERTEX_ARRAY );
		glDrawArrays( GL_TRIANGLES, 0, 3 );
		glDisableClientState( GL_VERTEX_ARRAY );
	}

	glLoadIdentity();
	glDisable( GL_BLEND );
}

gi_uie *gi_uie_create( gi_ui_base *base, uint type,
		f32 pos_x, f32 pos_y,
		f32 width, f32 height,
		const char *text, gi_uie *last,
		gi_uie_func func, uint flags )
{
	gi_uie *uie = malloc( sizeof( gi_uie ) );

	uint i, count = 0;
	uie->base = base;
	uie->last = last;
	uie->next_list = NULL;
	uie->type = type;
	uie->x = pos_x;
	uie->y = pos_y;
	uie->w = width;
	uie->h = height;
	uie->func = func;
	uie->vdata = NULL;
	uie->flags = flags;
	uie->info = NULL;
	f32 _w, text_w = 0.0f;

	switch ( uie->type )
	{
		case GI_UIE_WINDOW:
		case GI_UIE_BUTTON:
		case GI_UIE_TEXT:
		case GI_UIE_EDIT:
		case GI_UIE_CHECK:
			uie->text.string = gi_string_create( base->font, base->fdefsize, text );
			memcpy( &uie->text.string->fcl,
					base->cltheme[ GI_UI_COLOR_FONT ], 4 );
			text_w = gi_string_wcalc( uie->text.string );
			break;

		case GI_UIE_COMBO:
		case GI_UIE_LIST:
			while ( ( ( const char ** ) text ) [ count ] != NULL ) count++;
			uie->text.list = malloc( ( count + 1 ) * sizeof( gi_string* ) );
			for ( i = 0; i < count; i++ ) {
				uie->text.list[ i ] = gi_string_create( base->font, base->fdefsize,
						( ( const char ** ) text ) [ i ] );
				memcpy( &uie->text.list[ i ]->fcl,
						base->cltheme[ GI_UI_COLOR_FONT ], 4 );
				_w = gi_string_wcalc( uie->text.list[ i ] );
				text_w = text_w < _w ? _w : text_w ;
			}
			uie->text.list[ count ] = NULL;
			break;

		case GI_UIE_IMAGE:
			uie->text.id = *( uint* ) text;
			break;
	}

	if ( uie->h < ( uie->base->fdefsize * 1.5f ) )
		uie->h = uie->base->fdefsize * 1.5f;
	if ( uie->w < ( text_w + uie->base->fdefsize * 0.5f ) )
		uie->w = text_w + uie->base->fdefsize * 0.5f;

	uint size ;
	switch ( uie->type )
	{
		case GI_UIE_WINDOW:
			uie->vdata = malloc( 8 * ( 8 ) );
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 0 + 0 ) = 0.0f;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 0 + 4 ) = 0.0f;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 1 + 0 ) = 0.0f;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 1 + 4 ) = uie->h;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 2 + 0 ) = uie->w;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 2 + 4 ) = uie->h;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 3 + 0 ) = uie->w;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 3 + 4 ) = 0.0f;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 4 + 0 ) = 0.0f;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 4 + 4 ) = 0.0f;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 5 + 0 ) = 0.0f;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 5 + 4 ) = 
				uie->base->fdefsize * 1.5f;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 6 + 0 ) = uie->w;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 6 + 4 ) = 
				uie->base->fdefsize * 1.5f;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 7 + 0 ) = uie->w;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 7 + 4 ) = 0.0f;
			uie->tx = ( uie->w - text_w ) * 0.5f;
			uie->ty = uie->base->fdefsize * 0.25f;
			uie->flags |= GI_UIE_NOSELECT;
			break;

		case GI_UIE_BUTTON:
			uie->vdata = malloc( 4 * ( 8 ) );
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 0 + 0 ) = 0.0f;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 0 + 4 ) = 0.0f;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 1 + 0 ) = 0.0f;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 1 + 4 ) = uie->h;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 2 + 0 ) = uie->w;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 2 + 4 ) = uie->h;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 3 + 0 ) = uie->w;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 3 + 4 ) = 0.0f;
			uie->tx = ( uie->w - text_w ) * 0.5f;
			uie->ty = ( uie->h - uie->base->fdefsize ) * 0.5f;
			uie->flags |= GI_UIE_PUSHABLE;
			uie->info = malloc( sizeof( gi_uie_button ) );
			memset( uie->info, FALSE, sizeof( gi_uie_button ) );
			break;

		case GI_UIE_TEXT:
			uie->tx = 0.0f;
			uie->ty = 0.0f;
			uie->flags |= GI_UIE_NOSELECT;
			break;

		case GI_UIE_EDIT:
			break;

		case GI_UIE_CHECK:
			uie->w += uie->h;
			uie->vdata = malloc( 8 * ( 8 ) );
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 0 + 0 ) = 0.0f;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 0 + 4 ) = 0.0f;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 1 + 0 ) = 0.0f;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 1 + 4 ) = uie->h;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 2 + 0 ) = uie->w;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 2 + 4 ) = uie->h;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 3 + 0 ) = uie->w;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 3 + 4 ) = 0.0f;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 4 + 0 ) = uie->h * 0.33f;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 4 + 4 ) = uie->h * 0.33f;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 5 + 0 ) = uie->h * 0.33f;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 5 + 4 ) = uie->h * 0.66f;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 6 + 0 ) = uie->h * 0.66f;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 6 + 4 ) = uie->h * 0.66f;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 7 + 0 ) = uie->h * 0.66f;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 7 + 4 ) = uie->h * 0.33f;
			uie->tx = uie->h ;
			uie->ty = ( uie->h - uie->base->fdefsize ) * 0.5f;
			uie->flags |= GI_UIE_PUSHABLE;
			uie->info = malloc( sizeof( gi_uie_check ) );
			memset( uie->info, FALSE, sizeof( gi_uie_check ) );
			break;

		case GI_UIE_COMBO:
			uie->w += uie->h * 2.0f;
			uie->vdata = malloc( 10 * ( 8 ) );
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 0 + 0 ) = 0.0f;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 0 + 4 ) = 0.0f;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 1 + 0 ) = 0.0f;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 1 + 4 ) = uie->h;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 2 + 0 ) = uie->w;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 2 + 4 ) = uie->h;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 3 + 0 ) = uie->w;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 3 + 4 ) = 0.0f;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 4 + 0 ) = uie->h * 0.8f;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 4 + 4 ) = uie->h * 0.2f;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 5 + 0 ) = uie->h * 0.3f;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 5 + 4 ) = uie->h * 0.5f;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 6 + 0 ) = uie->h * 0.8f;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 6 + 4 ) = uie->h * 0.8f;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 7 + 0 ) = uie->w - uie->h * 0.3f;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 7 + 4 ) = uie->h * 0.5f;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 8 + 0 ) = uie->w - uie->h * 0.8f;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 8 + 4 ) = uie->h * 0.2f;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 9 + 0 ) = uie->w - uie->h * 0.8f;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 9 + 4 ) = uie->h * 0.8f;
			uie->tx = uie->h ;
			uie->ty = ( uie->h - uie->base->fdefsize ) * 0.5f;
			uie->flags |= GI_UIE_PUSHABLE;
			uie->info = malloc( sizeof( gi_uie_combo ) );
			GIUI_COMBO_INFO( uie )->index = 1;
			GIUI_COMBO_INFO( uie )->length = count;
			GIUI_COMBO_INFO( uie )->light = FALSE;
			GIUI_COMBO_INFO( uie )->l_push = FALSE;
			GIUI_COMBO_INFO( uie )->r_push = FALSE;
			break;

		case GI_UIE_LIST:
			size = ( uint ) ( uie->h / ( uie->base->fdefsize * 1.5f ) );
			size = count < size ? count : size ;
			uie->h = ( uie->base->fdefsize * 1.5f ) * ( f32 ) size;
			uie->vdata = malloc( 8 * ( 8 ) );
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 0 + 0 ) = 0.0f;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 0 + 4 ) = 0.0f;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 1 + 0 ) = 0.0f;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 1 + 4 ) = uie->h;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 2 + 0 ) = uie->w;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 2 + 4 ) = uie->h;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 3 + 0 ) = uie->w;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 3 + 4 ) = 0.0f;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 4 + 0 ) = 0.0f;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 4 + 4 ) = 0.0f;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 5 + 0 ) = 0.0f;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 5 + 4 ) = 
				uie->base->fdefsize * 1.5f;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 6 + 0 ) = uie->w;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 6 + 4 ) = 
				uie->base->fdefsize * 1.5f;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 7 + 0 ) = uie->w;
			* ( f32* ) ( ( uint ) uie->vdata + 8 * 7 + 4 ) = 0.0f;
			uie->tx = uie->base->fdefsize * 0.25f;
			uie->ty = uie->base->fdefsize * 0.25f;
			uie->flags |= GI_UIE_PUSHABLE;
			uie->info = malloc( sizeof( gi_uie_list ) );
			GIUI_LIST_INFO( uie )->light = 0;
			GIUI_LIST_INFO( uie )->index = 0;
			GIUI_LIST_INFO( uie )->scroll = 0;
			GIUI_LIST_INFO( uie )->length = count;
			GIUI_LIST_INFO( uie )->size = size;
			GIUI_LIST_INFO( uie )->push = FALSE;
			break;
	}

	if ( uie->last == NULL )
	{
		count = 0;
		while ( uie->base->root_uie_list[ count ] != NULL ) count++;
		if ( count < 7 )
		{
			uie->base->root_uie_list[ count ] = ( void* ) uie;
			uie->base->root_uie_list[ count + 1 ] = NULL;
		}
		else
			fprintf( stderr, "%s:%i [ERROR] too many root UIEs \n",
					__FILE__, __LINE__ );
	}
	else
	{
		gi_uie *_last = ( gi_uie* ) uie->last;

		if ( _last->next_list == NULL )
		{
			_last->next_list = malloc( 2 * sizeof( gi_uie* ) );
			_last->next_list[ 0 ] = uie;
			_last->next_list[ 1 ] = NULL;
		}
		else
		{
			count = 0;
			while ( _last->next_list[ count ] != NULL ) count++;
			_last->next_list = realloc( _last->next_list,
					( count + 2 ) * sizeof( gi_uie* ) );
			_last->next_list[ count ] = uie;
			_last->next_list[ count + 1 ] = NULL;
		}
	}

	return uie;
}

void gi_uie_mfly( gi_uie *uie, f32 x, f32 y, f32 time )
{
	gi_uie_fly *list = ( gi_uie_fly* ) uie->base->fly_uie_list;

	uint i = 0 ;
	while ( ( list[ i ].use == TRUE ) &&
			( i < 3 ) ) i++;

	list[ i ].target = uie;
	list[ i ].src_x = uie->x;
	list[ i ].src_y = uie->y;
	list[ i ].dest_x = x;
	list[ i ].dest_y = y;
	list[ i ].pos = 1.0f;
	list[ i ].acc = 2.0f / ( time * time );
	list[ i ].time = time;
	list[ i ].use = TRUE;

	uie->base->fly = TRUE;
}

void gi_uie_select( gi_uie *uie )
{
	if ( uie->flags & GI_UIE_NOSELECT )
	{
		gi_uie **next = uie->next_list;
		if ( next != NULL )
		{
			int i = 0;
			while ( next[ i ] != NULL )
				if ( ( next[ i ]->flags & GI_UIE_NOSELECT ) == 0 )
				{
					uie->base->s_uie = next[ i ];
					return ;
				}
				else
					i++;
		}
	}
	else
		uie->base->s_uie = uie;
}

void gi_uie_delete( gi_uie *uie )
{
	uint i = 0;
	if ( uie->next_list != NULL )
	{
		while ( uie->next_list[ i ] != NULL )
			gi_uie_delete( uie->next_list[ i++ ] );
		free( uie->next_list );
	}

	i = 0 ;
	if ( uie->last == NULL )
	{
		gi_uie **list = ( gi_uie** ) uie->base->root_uie_list;
		while ( list[ i ] != uie ) i++;
		list[ i ] = NULL;
		do
			list[ i ] = list[ i + 1 ];
		while ( list[ i++ ] != NULL ) ;
	}
	else
	{
		gi_uie **list = ( gi_uie** ) ( ( gi_uie* ) uie->last ) -> next_list;
		while ( list[ i ] != uie ) i++;
		list[ i ] = NULL;
		do
			list[ i ] = list[ i + 1 ];
		while ( list[ i++ ] != NULL ) ;
	}

	switch ( uie->type )
	{
		case GI_UIE_WINDOW:
		case GI_UIE_BUTTON:
		case GI_UIE_TEXT:
		case GI_UIE_EDIT:
		case GI_UIE_CHECK:
			gi_string_delete( uie->text.string );
			break;

		case GI_UIE_COMBO:
		case GI_UIE_LIST:
			i = 0;
			while ( uie->text.list[ i ] != NULL )
				gi_string_delete( uie->text.list[ i++ ] );
			free( uie->text.list );
			break;

		case GI_UIE_IMAGE:
			break;
	}

	if ( uie->info != NULL )
		free( uie->info );
	if ( uie->vdata != NULL )
		free( uie->vdata );
	free( uie );
}

