#ifndef _H_GIUI_
#define _H_GIUI_

#include "gifont.h"
#include "gint.h"

#define GIUI_UIE(UIE)		( ( gi_uie* ) UIE )
#define GIUI_BUTTON_INFO(UIE)	( ( gi_uie_button* ) ( GIUI_UIE( UIE ) )->info )
#define GIUI_CHECK_INFO(UIE)	( ( gi_uie_check* ) ( GIUI_UIE( UIE ) )->info )
#define GIUI_COMBO_INFO(UIE)	( ( gi_uie_combo* ) ( GIUI_UIE( UIE ) )->info )
#define GIUI_LIST_INFO(UIE)	( ( gi_uie_list* ) ( GIUI_UIE( UIE ) )->info )

enum {
	GI_UIE_WINDOW		= 0x0000,
	GI_UIE_BUTTON,
	GI_UIE_TEXT,
	GI_UIE_EDIT,
	GI_UIE_CHECK,
	GI_UIE_COMBO,
	GI_UIE_LIST,
	GI_UIE_IMAGE
};

enum {
	GI_UIE_HIDDEN		= 0x0001,
	GI_UIE_MOVEABLE		= 0x0002,
	GI_UIE_PUSHABLE		= 0x0004,
	GI_UIE_BORDERED		= 0x0008,
	GI_UIE_NOSELECT		= 0x0010
};

enum {
	GI_UIE_PUSH		= 0x0000,
	GI_UIE_UNPUSH,
	GI_UIE_MOVE,
	GI_UIE_UNMOVE,
	GI_UIE_LIGHT,
	GI_UIE_UNLIGHT,
	GI_UIE_SCROLL
};

typedef struct gi_ui_base gi_ui_base;
typedef struct gi_uie gi_uie;
typedef struct gi_uie_fly gi_uie_fly;

typedef void ( *gi_uie_func ) ( gi_uie *uie, uint action );

struct gi_ui_base {
	gi_font *font;			/* main font */
	f32 wnd_w, wnd_h;		/* target window sizes */
	f32 upt_x, upt_y;		/* cursor position */
	f32 fdefsize;			/* default font size */
	ui8 cltheme[ 10 ][ 4 ];		/* color theme */
	void *vupt;			/* cursor vertexes */
	void *vsel;			/* selector vertexes */
	gi_uie *l_uie;			/* lighting UIE */
	gi_uie *s_uie;			/* selected UIE */
	f32 move_upt_x;			/* initial cursor position on X */
	f32 move_upt_y;			/* initial cursor position on Y */
	f32 move_uie_x;			/* initial UIE position on X */
	f32 move_uie_y;			/* initial UIE position on Y */
	bool move;			/* move lighting UIE */
	bool fly;			/* enable fly */
	bool upt_on;			/* visible cursor */
	bool lref;			/* light refresh */
	bool glref;			/* need OpenGL redraw */
	gi_uie_fly *fly_uie_list;	/* list of flying UIEs */
	gi_uie **root_uie_list;		/* list of root UIEs */
};

struct gi_uie {
	gi_ui_base *base;		/* base structure */
	gi_uie *last;			/* parent UIE */
	gi_uie **next_list;		/* next UIEs */
	uint type;			/* type of this UIE */
	
	f32 x, y;			/* UIE position */
	f32 w, h;			/* UIE sizes */
	union {
		gi_string	**list;
		gi_string	*string;
		uint		id;
	} text;				/* UIE text */
	f32 tx, ty;			/* position of text */
	gi_uie_func func;		/* callback function */
	void *vdata;			/* UIE vertexes */
	uint flags;			/* UIE flags */
	
	uint label;			/* label for any targets */
	void *info;			/* additional info */
	
};	/* User Interface Element */

typedef struct {
	bool light;			/* cursor on UIE */
	bool push;			/* UIE is pushed */
} gi_uie_button;

typedef struct {
	bool light;			/* cursor on UIE */
	bool push;			/* UIE is pushed */
	bool check;			/* UIE is ckecked */
} gi_uie_check;

typedef struct {
	uint index;			/* index of active element */
	uint length;			/* length of list */
	bool light;			/* cursor on UIE */
	bool l_push;			/* left is pushed */
	bool r_push;			/* right is pushed */
} gi_uie_combo;

typedef struct {
	uint light;			/* index of light element */
	uint index;			/* index of active element */
	uint scroll;			/* scroll offset */
	uint length;			/* length of list */
	uint size;			/* visible size of list */
	bool push;			/* UIE is pushed */
} gi_uie_list;

struct gi_uie_fly {
	gi_uie *target;			/* target UIE */
	f32 src_x;			/* source position on X */
	f32 src_y;			/* source position on Y */
	f32 dest_x;			/* destination position on X */
	f32 dest_y;			/* destination position on Y */
	f32 pos;			/* position 0-1 value */
	f32 acc;			/* acceleration */
	f32 time;			/* animation time */
	bool use;			/* using this element */
};	/* flying UIEs */

gi_ui_base *gi_ui_base_create( gi_font *font );
void gi_ui_base_delete( gi_ui_base *base );
void gi_ui_update( gi_ui_base *base, gi_input *input, f64 dt );
void gi_ui_draw( gi_ui_base *base );

gi_uie *gi_uie_create( gi_ui_base *base, uint type,
			f32 pos_x, f32 pos_y,
			f32 width, f32 height,
			const char *text, gi_uie *last,
			gi_uie_func func, uint flags );

void gi_uie_mfly( gi_uie *uie, f32 x, f32 y, f32 time );
void gi_uie_text( gi_uie *uie, const char *text );
void gi_uie_select( gi_uie *uie );
void gi_uie_delete( gi_uie *uie );

#endif /* _H_GIUI_ */

