#define _XOPEN_SOURCE 500

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if defined(_GI_X11)
# include <unistd.h>
#elif defined(_GI_WIN)
# include <windows.h>
#endif

#include "def.h"
#include "game.h"
#include "sys/config.h"
#include "gint/gint.h"

static bool run = TRUE, game = FALSE;
static gi_uie *gi_menu, *gi_set, *gi_info;

static s_config *config;
static g_world *world;

static int set_level;

static void ui_quit( gi_uie *uie, uint action )
{
	switch ( action )
	{
		case GI_UIE_UNPUSH:
			run = FALSE;
			gi_uie_mfly( gi_menu, -0.4f, 0.4f, 0.8f );
			gi_uie_mfly( gi_set, 0.2f, -0.8f, 0.8f );
			gi_uie_mfly( gi_info, 0.2f, -0.8f, 0.8f );
			break;
	}
}

static void ui_mv_set( gi_uie *uie, uint action )
{
	switch ( action )
	{
		case GI_UIE_UNPUSH:
			gi_uie_mfly( gi_menu, -0.4f, 0.4f, 0.8f );
			gi_uie_mfly( gi_set, 0.2f, 0.2f, 0.8f );
			gi_uie_select( gi_set );
			break;
	}
}

static void ui_mv_info( gi_uie *uie, uint action )
{
	switch ( action )
	{
		case GI_UIE_UNPUSH:
			gi_uie_mfly( gi_menu, -0.4f, 0.4f, 0.8f );
			gi_uie_mfly( gi_info, 0.2f, 0.2f, 0.8f );
			gi_uie_select( gi_info );
			break;
	}
}

static void ui_mv_menu( gi_uie *uie, uint action )
{
	switch ( action )
	{
		case GI_UIE_UNPUSH:
			gi_uie_mfly( gi_menu, 0.01f, 0.4f, 0.8f );
			gi_uie_select( gi_menu );
			if ( GIUI_UIE( uie )->label == 2 )
				gi_uie_mfly( gi_set, 0.2f, -0.8f, 0.8f );
			else if ( GIUI_UIE( uie )->label == 3 )
				gi_uie_mfly( gi_info, 0.2f, -0.8f, 0.8f );
			break;
	}
}

static void ui_game_start( gi_uie *uie, uint action )
{
	switch ( action )
	{
		case GI_UIE_UNPUSH:
			world = g_world_create( uie->base->font, SHARE_PATH, set_level );
			game = TRUE;
			break;
	}
}

static void ui_game_stop( gi_uie *uie, uint action )
{
	switch ( action )
	{
		case GI_UIE_UNPUSH:
			g_world_delete( world );
			game = FALSE;
			break;
	}
}

static void ui_set_val( gi_uie *uie, uint action )
{
	switch ( action )
	{
		case GI_UIE_UNPUSH:
			switch( uie->label )
			{
				case 1:
					set_level = GIUI_LIST_INFO( uie )->index - 1;
					break;

				case 3:
					if ( GIUI_CHECK_INFO( uie )->check )
						glEnable( GL_FOG );
					else
						glDisable( GL_FOG );
					break;
			}
			break;
	}
}

int main( int argc, char* argv[] )
{
	gi_create();

	config = s_config_read( SHARE_PATH "ball.conf" );
	if ( config == NULL )
		return EXIT_FAILURE;

	gw_params params = {
		.cbits = s_config_get_int( config, "cbits" ),
		.zbits = s_config_get_int( config, "zbits" ),
		.sbits = 0,
		.flags = ( s_config_get_int( config, "fullscreen" ) ) ? GW_FULLSCREEN : 0
	};

	int width = s_config_get_int( config, "width" );
	int height = s_config_get_int( config, "height" );

	params.flags |= GW_HIDECURSOR;

	gw_create( PROGRAM_TITLE,
			0, 0,
			width, height,
			&params );

	glHint( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );
	glShadeModel( GL_SMOOTH );

	glEnable( GL_DEPTH_TEST );
	glDepthFunc( GL_LEQUAL );

	glEnable( GL_CULL_FACE );
	glCullFace( GL_BACK );

	glEnable( GL_COLOR_MATERIAL );
	glEnable( GL_TEXTURE_2D );

	f32 cfog[] = { 0.0f, 0.0f, 0.0f, 1.0f };

	glEnable( GL_FOG );
	glFogi( GL_FOG_MODE, GL_LINEAR );
	glFogfv( GL_FOG_COLOR, cfog );
	glFogf( GL_FOG_START, 1.0f );
	glFogf( GL_FOG_END, 40.0f );

	// set
	set_level = 0;

	gi_font *fx16 = gi_font_create( 0.6f, SHARE_PATH "font_16px.png" );
	gi_ui_base *gui = gi_ui_base_create( fx16 );

	gi_menu = gi_uie_create( gui, GI_UIE_WINDOW,
			-0.4f, 0.4f,
			gui->fdefsize * 8.0f,
			gui->fdefsize * 11.0f,
			_("main menu"), NULL, NULL, GI_UIE_MOVEABLE | GI_UIE_BORDERED );

	gi_uie *bt = gi_uie_create( gui, GI_UIE_BUTTON,
			gui->fdefsize, gui->fdefsize * 2.0f, 0.0f, 0.0f,
			_("start game"), gi_menu, ui_mv_set, 0 );
	gi_uie_create( gui, GI_UIE_BUTTON,
			gui->fdefsize, gui->fdefsize * 4.0f, bt->w, 0.0f,
			_("info"), gi_menu, ui_mv_info, 0 );
	gi_uie *qu = gi_uie_create( gui, GI_UIE_BUTTON,
			gui->fdefsize, gui->fdefsize * 8.0f, bt->w, 0.0f,
			_("quit"), gi_menu, ui_quit, 0 );
	GIUI_UIE( qu )->label = 12 ;

	gi_set = gi_uie_create( gui, GI_UIE_WINDOW,
			0.2f, -0.8f, 0.7f, 0.6f,
			_("game options"), NULL, NULL, GI_UIE_MOVEABLE | GI_UIE_BORDERED );

	char *level_list[] = { _("Ice Puppy"), _("Four Lines"), _("Hard Ground"),
		"","","","", NULL } ;
	gi_uie_create( gui, GI_UIE_TEXT,
			gui->fdefsize, gui->fdefsize * 2.0f, 0.0f, 0.0f,
			_("level"), gi_set, NULL, 0 );
	gi_uie *lev = gi_uie_create( gui, GI_UIE_LIST,
			gui->fdefsize + 0.3f, gui->fdefsize * 2.0f, 0.35f, 0.2f,
			( char* ) level_list, gi_set, ui_set_val, 0 );
	GIUI_UIE( lev )->label = 1 ;
	GIUI_LIST_INFO( lev )->index = set_level + 1;

	char *dif_list[] = { _("easy"), _("medium"), _("hard"), NULL } ;
	gi_uie_create( gui, GI_UIE_TEXT,
			gui->fdefsize, gui->fdefsize * 12.0f, 0.0f, 0.0f,
			_("difficulty"), gi_set, NULL, 0 );
	gi_uie *dif = gi_uie_create( gui, GI_UIE_COMBO,
			gui->fdefsize + 0.3f, gui->fdefsize * 12.0f, 0.0f, 0.0f,
			( char* ) dif_list, gi_set, ui_set_val, 0 );
	GIUI_UIE( dif )->label = 2 ;

	gi_uie *fo = gi_uie_create( gui, GI_UIE_CHECK,
			gui->fdefsize + 0.3f, gui->fdefsize * 14.0f, 0.0f, 0.0f,
			_("fog"), gi_set, ui_set_val, 0 );
	GIUI_UIE( fo )->label = 3 ;
	GIUI_CHECK_INFO( fo )->check = TRUE;

	gi_uie *bk = gi_uie_create( gui, GI_UIE_BUTTON,
			gui->fdefsize, 0.6f - gui->fdefsize * 3.0f, 0.0f, 0.0f,
			_("back    "), gi_set, ui_mv_menu, 0 );
	GIUI_UIE( bk )->label = 2 ;

	bk = gi_uie_create( gui, GI_UIE_BUTTON,
			gui->fdefsize * 2.0f + bk->w, 0.6f - gui->fdefsize * 3.0f, 0.0f, 0.0f,
			_("play    "), gi_set, ui_game_start, 0 );
	GIUI_UIE( bk )->label = 3 ;

	gi_info = gi_uie_create( gui, GI_UIE_WINDOW,
			0.2f, -0.8f, 0.7f, 0.6f,
			_("information"), NULL, NULL, GI_UIE_MOVEABLE | GI_UIE_BORDERED );
	gi_uie_create( gui, GI_UIE_TEXT,
			gui->fdefsize, gui->fdefsize * 2.0f, 0.0f, 0.0f,
			_("Independed Games Developers Contests "),
			gi_info, NULL, 0 );
	gi_uie_create( gui, GI_UIE_TEXT,
			gui->fdefsize, gui->fdefsize * 4.0f, 0.0f, 0.0f,
			_("        <http://igdc.ru/>"),
			gi_info, NULL, 0 );
	gi_uie_create( gui, GI_UIE_TEXT,
			gui->fdefsize, gui->fdefsize * 8.0f, 0.0f, 0.0f,
			_("This is free software, ... "),
			gi_info, NULL, 0 );
	gi_uie_create( gui, GI_UIE_TEXT,
			gui->fdefsize, gui->fdefsize * 18.0f, 0.0f, 0.0f,
			_("NE_Mashinist <nemashinist@rambler.ru>"),
			gi_info, NULL, 0 );
	bk = gi_uie_create( gui, GI_UIE_BUTTON,
			gui->fdefsize, 0.6f - gui->fdefsize * 3.0f, 0.0f, 0.0f,
			_("back    "), gi_info, ui_mv_menu, 0 );
	GIUI_UIE( bk )->label = 3 ;

	gi_uie_mfly( gi_menu, 0.01f, 0.4f, 0.8f );
	gi_uie_select( gi_menu );

	gi_string *title = gi_string_create( fx16, 0.1f, _("Ball Dest") );
	title->fcl[ 0 ] = 255;
	title->fcl[ 1 ] = 32;
	title->fcl[ 2 ] = 32;

	ui64 t0, t1 = t0 = gi_timer_ui64();

	while ( ( run ) || ( gui->fly ) )
	{
		t1 = t0;
		t0 = gi_timer_ui64();
		f64 dt = ( f64 ) ( t0 - t1 ) * 0.000001;

		gi_input *input = gi_input_update();

		if ( ( input->keys[ GI_ESC ] == GI_KEY_DOWN ) ||
				( input->keys[ GI_Q ] == GI_KEY_DOWN ) )
		{
			if ( game )
			{
				ui_game_stop( qu, GI_UIE_UNPUSH );
				input->keys[ GI_ESC ] = GI_KEY_OFF;
				input->keys[ GI_Q ] = GI_KEY_OFF;
			}
			else
				ui_quit( qu, GI_UIE_UNPUSH );
		}

		if ( game )
			g_world_update( world, input, dt );
		else
			gi_ui_update( gui, input, dt );

		glViewport( 0, 0, width, height );
		glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

		if ( game )
		{
			gi_perspective( 65.0f,
					( ( f32 ) width ) / ( ( f32 ) height ),
					0.5f, 40.0f );

			g_world_draw( world );
		}
		else
		{
			gi_ortho();

			glEnable( GL_BLEND );
			glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
			gi_string_draw_xy( title, 0.12f, 0.1f );
			glDisable( GL_BLEND );

			gi_ui_draw( gui );
		}

		gw_swap_buffers();

#if defined(_GI_X11)
		usleep( 1000 );		/* 1k fps limit */
#elif defined(_GI_WIN)
		Sleep( 1 );
#endif
	}

	gi_string_delete( title );

	gi_ui_base_delete( gui );
	gi_font_delete( fx16 );
	s_config_delete( config );

	gw_delete();
	gi_delete();

	return EXIT_SUCCESS;
}

