#define _XOPEN_SOURCE 500

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "../def.h"
#include "stack.h"
#include "map.h"

s_map *s_map_create( s_map_compare f )
{
	s_map *map = malloc( sizeof( s_map ) );

	map->root = NULL;
	map->func = f;

	return ( map );
}

void s_map_delete( s_map *map )
{
	s_stack *stack = s_stack_create();
	s_map_node *node = s_map_get_right( map );
	while ( node != NULL )
	{
		s_stack_push( stack, node );
		node = s_map_move_left( map, node );
	}
	while ( ( node = s_stack_pop( stack ) ) != NULL )
		free( node );
	free( map );
	s_stack_delete( stack );
}

void *s_map_set( s_map *map, const void *index, void *data )
{
	int cmp = 0;
	s_map_node *top_node = NULL, *curr = map->root;

	while ( curr != NULL )
	{
		cmp = map->func( curr->index, index );
		top_node = curr;

		if ( cmp < 0 )
			curr = curr->right;
		else
			if ( cmp == 0 )
			{
				void *old = curr->data;
				curr->data = data;
				return ( old );
			}
			else
				curr = curr->left;
	}

	s_map_node *new = malloc( sizeof( s_map_node ) );
	new->index = index;
	new->data = data;
	new->left = NULL;
	new->right = NULL;
	new->top = top_node;
	new->sign = 0;

	if ( top_node != NULL )
	{
		if ( cmp < 0 )
			top_node->right = new;
		else
			top_node->left = new;
	}
	else
		map->root = new;

	//FIXME: tree need for balance
	
	return NULL;
}

s_map_node *s_map_get( s_map *map, const void *index )
{
	int cmp ;
	s_map_node *curr = map->root;

	while ( curr != NULL )
	{
		cmp = map->func( curr->index, index );

		if ( cmp < 0 )
			curr = curr->right;
		else
			if ( cmp == 0 )
				return ( curr );
			else
				curr = curr->left;
	}

	return NULL;
}

s_map_node *s_map_get_left( s_map *map )
{
	s_map_node *curr = map->root;
	while ( curr->left != NULL )
		curr = curr->left;
	return ( curr );
}

s_map_node *s_map_get_right( s_map *map )
{
	s_map_node *curr = map->root;
	while ( curr->right != NULL )
		curr = curr->right;
	return ( curr );
}

s_map_node *s_map_move_left( s_map *map, s_map_node *node )
{
	s_map_node *curr = node, *top;

	if ( curr->left != NULL )
	{
		curr = curr->left;
		while ( curr->right != NULL )
			curr = curr->right;
		return ( curr );
	}
	else 
		if ( curr->top != NULL )
		{
			top = curr->top;
			while ( top->left == curr )
			{
				curr = top;
				top = top->top;
				if ( top == NULL )
					return NULL;
			}
			return ( top );
		}

	return NULL;
}

s_map_node *s_map_move_right( s_map *map, s_map_node *node )
{
	s_map_node *curr = node, *top;

	if ( curr->right != NULL )
	{
		curr = curr->right;
		while ( curr->left != NULL )
			curr = curr->left;
		return ( curr );
	}
	else 
		if ( curr->top != NULL )
		{
			top = curr->top;
			while ( top->right == curr )
			{
				curr = top;
				top = top->top;
				if ( top == NULL )
					return NULL;
			}
			return ( top );
		}

	return NULL;
}

