#!BPY

"""
Name: 'gint (.mesh)...'
Blender: 243
Group: 'Export'
Tooltip: 'Export to .mesh file'
"""

__author__ = 'amaora'
__version__ = '1.0.2'
__email__ = 'nemashinist@rambler.ru'
__bpydoc__ = """\
"""

import Blender
import struct

class gint_mesh:
	def __init__ ( self ):
		self.idata = []
		self.vdata = []

	def add_vert( self, co, uv, no ):
		self.idata.append( len( self.idata ) )
		self.vdata.append( [	co[ 0 ], co[ 1 ], co[ 2 ],	\
					uv[ 0 ], uv[ 1 ],		\
					no[ 0 ], no[ 1 ], no[ 2 ] ] )

	def write( self, file ):
		fd = open( file, 'wb' )
		fd.write( struct.pack( '<H', len( self.idata ) ) )
		fd.write( struct.pack( '<H', len( self.vdata ) ) )
		for indx in self.idata:
			fd.write( struct.pack( '<H', indx ) )
		for vert in self.vdata:
			fd.write( struct.pack( '<f', vert[ 0 ] ) )
			fd.write( struct.pack( '<f', vert[ 1 ] ) )
			fd.write( struct.pack( '<f', vert[ 2 ] ) )
			fd.write( struct.pack( '<f', vert[ 3 ] ) )
			fd.write( struct.pack( '<f', vert[ 4 ] ) )
			fd.write( struct.pack( '<b', vert[ 5 ] * 127.0 ) )
			fd.write( struct.pack( '<b', vert[ 6 ] * 127.0 ) )
			fd.write( struct.pack( '<b', vert[ 7 ] * 127.0 ) )
			fd.write( struct.pack( '<b', 0 ) )
		fd.close()

	def pack( self, mat, mesh ):
		rot = mat.rotationPart()
		for face in mesh.faces:
			i = 0
			for vert in face.v:
				self.add_vert( mat * vert.co, face.uv[ i ], rot * vert.no )
				if i > 2 :
					self.add_vert(	mat * face.v[ 0 ].co,		\
							face.uv[ 0 ],		\
							rot * face.v[ 0 ].no )
					self.add_vert(	mat * face.v[ i - 1 ].co,	\
							face.uv[ i - 1 ],	\
							rot * face.v[ i - 1 ].no )
				i += 1

def export_mesh( file ):
	gmesh = gint_mesh()
	objs = Blender.Scene.GetCurrent().objects
	for obj in objs:
		if obj.getType() == 'Mesh':
			mesh = obj.getData( mesh = 1 )
			mat = obj.getMatrix()
			if mesh.faceUV == True:
				gmesh.pack( mat, mesh )
			else:
				print '*** mesh haven\'t uv coords'
				return
	gmesh.write( file )

Blender.Window.FileSelector( export_mesh, 'Export', Blender.sys.makename( ext = '.mesh' ) )
